#from morebs2.fit_2n2 import *
#from morebs2.random_generators import *
#from morebs2.deline_helpers import *
from .fit_2n2 import *
from .random_generators import *
from .deline_helpers import *
from collections import defaultdict,Counter

class Delineation:

    def __init__(self,label,clockwise,parentIds = [],childIds = [],idn = "0"):
        """
        data structure used to store and calculate points for delineation of a sequence
        of two-dimensional points

        :attribute d: container for point sequences corresponding to each direction (l,r,t,b)
        :type d: dict 
        :attribute d_: sequence of :class>`DCurves` corresponding to `d`. 
        :type d_: list 
        :attribute label: the classification of the points  
        :type label: int
        :attribute clockwise: is the delineation clockwise? 
        :type clockwise: bool
        """
        
        self.d = {}
        self.d_ = None
        self.label = label
        self.clockwise = clockwise
        self.parentIds = parentIds
        self.childIds = childIds
        return

    def add_edge(self,direction,edge):
        self.d[direction] = edge

    def no_duplicates(self):
        '''
        sequence correction algorithm by condition `no duplicates`. 
        '''
        # left and right
        self.d['l'],self.d['r'] = eliminate_duplicate_points(\
            self.d['l'],self.d['r'],0)

        # top and bottom
        self.d['t'],self.d['b'] = eliminate_duplicate_points(\
            self.d['t'],self.d['b'],1)

    def no_cross(self):
        '''
        sequence correction algorithm by condition `no cross`. 
        '''
        self.d['l'],self.d['r'] = eliminate_contrary_points(\
            self.d['l'],self.d['r'],1)

        self.d['t'],self.d['b'] = eliminate_contrary_points(\
            self.d['t'],self.d['b'],0)

        return

    def no_cross_on_axis(self,axis,maxNumberOfPoints=10):
        """
        NOTE: unnecessary loop, only one iteration will do. 
        """

        if axis == 1:
            k1,k2 = 'l','r'
        else:
            k1,k2 = 't','b'

        while True:
            l1,l2 = len(self.d[k1]), len(self.d[k2])
            self.d[k1],self.d[k2] = eliminate_contrary_points(\
                self.d[k1],self.d[k2],axis) 
            l1_,l2_ = len(self.d[k1]), len(self.d[k2])

            if l1_ == l1 and l2_ == l2:
                break

            if l1_ <= maxNumberOfPoints or l2_ <= maxNumberOfPoints:
                break

    def no_jags(self):
        '''
        sequence correction algorithm by condition `no jags`. 
        '''
        self.no_jags_on_edge('l')
        self.no_jags_on_edge('r')
        self.no_jags_on_edge('t')
        self.no_jags_on_edge('b')

    def no_jags_on_edge(self,direction,maxNumberOfPoints=10):
        assert direction in {'l','r','t','b'}

        while True: 
            l1 = len(self.d[direction])
            self.d[direction] = remove_jags_on_edges(self.d[direction],direction)
            l2 = len(self.d[direction])
            if l2 <= maxNumberOfPoints:
                break
            if l1 == l2: break
            
        return

    def draw_delineation(self):
        '''
        clockwise -> (t,r,b,l)
        counter-clockwise -> (t,l,b,r)
        '''
        self.d_ = []

        one = self.point_sequence_to_curve_set(self.d['t'],'t')
        
        if self.clockwise:
            two = self.point_sequence_to_curve_set(self.d['r'],'r')
            four = self.point_sequence_to_curve_set(self.d['l'],'l')
        else:
            four = self.point_sequence_to_curve_set(self.d['r'],'r')
            two = self.point_sequence_to_curve_set(self.d['l'],'l')

        three = self.point_sequence_to_curve_set(self.d['b'],'b')

        self.d_.extend(one)
        self.d_.extend(two)
        self.d_.extend(three)
        self.d_.extend(four)

    def point_pair_to_curve(self,p1,p2,ad):
        # case: line
        if p1[0] == p2[0] or\
            p1[1] == p2[1]:
            c = Line([p1,p2]) 
        else:
        # case: default is LogFit22
            direction = np.argsort(np.array([p1,p2])[:,1])
            c = LogFit22(np.array([p1,p2]),direction = direction)
        return DCurve(c,ad) 

    def point_sequence_to_curve_set(self,ps,ad):
        l = len(ps) - 1
        cs = []
        for i in range(0,l):
            cs.append(self.point_pair_to_curve(\
                deepcopy(ps[i]),deepcopy(ps[i+1]),ad))
        return cs

    def classify_point(self,p):
        '''
        :rtype: label if p in delineation, otherwise -1
        '''
        x1,x2,x3,x4 = self.point_to_relevant_curvepair(p)

        if x1 == None and x3 == None:
            return -1

        sc = self.classify_point_special_case(p,x1,x2,x3,x4)
        if sc != None:
            return self.label if sc else -1

        x1,x2 = (x1,x2) if type(x1) != None else (x3,x4)

        sc = self.classify_point_by_complementary_curves(p,x1,x2)
        return self.label if sc else -1

    def classify_point_special_case(self,p,c1,c2,c3,c4):
        '''
        provides a classification to point `p` by ('l','t') curves in
        the case where the value by the `l` and `r` OR `t` and `b` curves
        are equal. 

        NOTE: caution 
        '''

        if type(c1) != type(None): 
            p1,p2 = round(c1.x_given_y(p[1]),5),round(c2.x_given_y(p[1]),5)
            if p1 == p2:
                if type(c3) == type(None):
                    return None
                else:
                    return c1.is_ap(p) and c3.is_ap(p)
            return None
        
        if type(c3) != type(None): 
            p1,p2 = round(c3.y_given_x(p[0]),5),round(c4.y_given_x(p[0]),5)

            if p1 == p2:
                if type(c1) == type(None):
                    return None
                else:
                    return c1.is_ap(p) and c3.is_ap(p)
            return None
        return None

    def point_to_relevant_curvepair(self,p):
        '''
        searches for a curve pair (left&right)|(top&bottom)
        that 
        '''
        l,r,t,b = None,None,None,None

        for x in self.d_:
            if x.in_point_range(p):
                if x.ad == 'l': l = x
                elif x.ad == 'r': r = x
                elif x.ad == 't': t = x
                elif x.ad == 'b': b = x

        return l,r,t,b

    def classify_point_by_complementary_curves(self,p,c1,c2):

        s1 = c1.value_on_curve(p)
        s2 = c2.value_on_curve(p)

        if c1.ad in {'t','b'}:
            axis = 1
        else:
            axis = 0
        return p[axis] >= min([s1,s2]) and p[axis] <= max([s1,s2])

    def classify_point_by_curveset(self,p,rp,cs):
        for c in cs:
            if not c.in_point_range(p):
                continue

            if not self.classify_point_by_complementary_curves(p,rp,c):
                return -1

        return self.label

    def complement_set_to_curve(self,c):
        '''
        collects all curves in delineation with an 
        activation direction parallel to that of c
        that have a range that intersects that of c. 

        * note: method can only be called after `delineation_to_initial_estimate`. 
        '''
        axis = 1 if c.ad in {'l','r'} else 0
        pr = c.point_range()[axis]

        if c.ad == 'l':
            x = 'r'
        elif c.ad == 'r':
            x = 'l'
        elif c.ad == 'b':
            x = 't'
        else:
            x = 'b'

        q = []
        for c_ in self.d_:
            if c_.ad != x: continue
            pr2 = c_.point_range()[axis]
            if pr2[0] >= pr[0] and pr2[0] <= pr[1]:
                q.append(deepcopy(c_))
            elif pr2[1] >= pr[0] and pr2[1] <= pr[1]:
                q.append(deepcopy(c_))
        return q

    def visualize_delineation(self):
        ps = []
        colors = {'l': [1.,0.,0.,1.],\
            'r': [0.,1.,0.,1.],\
            't': [0.,0.,1.,1.],\
            'b': [0.,0.,0.,1.]
        }

        cs = []
        for c in self.d_:
            l = c.form_point_sequence(0.001)
            ps.extend(l)
            cs.extend([colors[c.ad] for j in range(len(l))])
        ps = np.array(ps)
        basic_2d_scatterplot(ps[:,0],ps[:,1],c=cs)
        return

    def is_proper_subset(self,d2):
        '''

        iterate through each curve, and check that their
        endpoints lay in the bounds of class:`Delineation` `d2`.

        NOTE: inefficient

        :return: bool, is instance in d2?
        '''

        for x in self.d_:
            p = x.get_point()
            for p_ in p:
                if d2.classify_point(p_) == -1:
                    return False
        return True

    @staticmethod
    def pertinent_curves_to_point(d,p):
        '''
        :return: list(curves with point `p` in its range)
        '''

        cs = []
        for c in d.d_:
            if c.in_point_range(p): cs.append(c)
        return cs 

class DLineate22:

    def __init__(self,xyl,clockwise=True,dmethod = "nocross",idn="0"):
        """
        Delineates a sequence of three-dimensional points using the :class:`Delineator`
        by one of the methods `nojag`,`nocross`, or `nodup`, each of which provides a
        different and possibly inaccurate delineation of the points (if data is "complex").

        Delineation targets the subset of points with the label of lowest frequency.

        :attribute xyl: three-dimensional points of (x,y,label). 
        :type xyl: np.ndarray 
        :attribute xyl_sorted_x: the `xyl` points sorted by x-axis. 
        :type xyl_sorted_x: nd.nparray 
        :attribute xyl_sorted_y: the `xyl` points sorted by y-axis. 
        :type xyl_sorted_y: nd.nparray  
        :attribute lc: label to label indices
        :type lc: dict  

        :attribute clockwise: is delineation clockwise?
        :type clockwise: bool
        :attribute label: target label of points to delineate
        :type label: int
        :attribute lpoints: target points to delineate
        :type lpoints: np.ndarray
        :attribute linfo: information of target points, [x-center,y-center,x-min,x-max,y-min,y-max]
        :type linfo: np.ndarray
        :attribute d: used to delineate the target points
        :type d: :class:`Delineator`
        :attribute dmethod: methodology used to correct points
        :type dmethod: string
        :attribute idn: identifier of class 
        :type idn: stringized int
        """


        assert xyl.shape[1] == 3, "invalid shape"
        assert dmethod in ["nodup","nojag","nocross"]
        
        self.xyl = np.round(xyl,5)
        self.xyl_sorted_x = None
        self.xyl_sorted_y = None 
        self.lc = None
        self.clockwise = clockwise
        self.label = None
        self.lpoints = None
        self.linfo = None
        self.d = None
        self.dmethod = dmethod
        self.idn = idn
        return

    ############# preprocessing methods

    def preprocess(self):
        self.label_counts()
        self.set_target_label()        
        self.sort_data() 
        
    def label_counts(self):
        self.lc = defaultdict(list)
        for (i,x) in enumerate(self.xyl):
            self.lc[x[2]].append(i)
        return

    def set_target_label(self,l=None):
        # default: choose min label
        if l == None:
            x = np.array([(k,len(v)) for (k,v) in self.lc.items()])
            a = np.argmin(x[:,1])
            self.label = x[a,0]
        else:
            self.label = l

        self.lpoints = deepcopy(self.xyl[self.lc[self.label]])
        self.lpoints = self.lpoints[:,:2]

    def target_label_analysis(self):
        c = self.calculate_center()
        ex = self.calculate_extremum()
        self.linfo = c + ex

    def calculate_center(self):
        '''
        calculates center for the lpoints
        '''
        return [np.mean(self.lpoints[:,0]),np.mean(self.lpoints[:,1])]

    def calculate_extremum(self):
        '''
        calculates extremum for the lpoints
        '''
        # x-min point
        xi = np.argmin(self.lpoints[:,0])
        # x-max point
        xa = np.argmax(self.lpoints[:,0])
        # y-min point
        yi = np.argmin(self.lpoints[:,1])
        # y-max point
        ya = np.argmax(self.lpoints[:,1])
        return [xi,xa,yi,ya]

    def sort_data(self):
        xs = np.argsort(self.lpoints[:,0])
        self.xyl_sorted_x = deepcopy(self.lpoints[xs]) #  deepcopy(self.xyl[xs])
        ys = np.argsort(self.lpoints[:,1])
        self.xyl_sorted_y = deepcopy(self.lpoints[ys]) #deepcopy(self.xyl[ys])
        return

    ################## initial delineation

    def collect_break_points(self,pi=[],ci=[]):
        '''
        clockwise -> [l + t -> increasing order, others in decreasing order]
        counter-clockwise -> [r + b -> increasing order, others in decreasing order]
        '''

        ds = ['l','r','t','b']
        self.d = Delineation(label=self.label,clockwise=self.clockwise,parentIds=pi,childIds=ci,idn =self.idn)
        for x in ds:
            edge = self.break_points_on_edge(x)
            self.d.add_edge(x,edge)

        if self.dmethod == "nodup":
            self.d.no_duplicates()
        elif self.dmethod == "nojag":
            self.d.no_jags()
        else:
            self.d.no_cross()

        self.assign_circular_directionality()
        self.d.draw_delineation()
        return

    def assign_circular_directionality(self):
        if self.clockwise:
            self.d.d['r'] = self.d.d['r'][::-1]
            self.d.d['b'] = self.d.d['b'][::-1]
        else:
            self.d.d['l'] = self.d.d['l'][::-1]
            self.d.d['t'] = self.d.d['t'][::-1]

    def break_points_on_edge(self,direction):
        '''
        direction := `l` is left edge,`r` is right edge,`t` is top edge,
                     `b` is bottom edge.

                     if `l`|`r`, output is ordered by axis 1. Otherwise, by axis 0. 
        '''
        edge = []

        stat = True
        i = None
        rd = self.xyl_sorted_y if direction in\
            {'l','r'} else self.xyl_sorted_x

        while stat:
            i = self.next_break_point(i,direction)
            if i == None: 
                stat = not stat
                continue
            edge.append(deepcopy(rd[i,:2]))
        return np.array(edge)

    def next_break_point(self,refi,direction):
        rd = None
        axis = None
        if direction in  {'l','r'}:
            rd = self.xyl_sorted_y
            axis = 1
        else:
            rd = self.xyl_sorted_x
            axis = 0
        
        j = None
        if refi == None:
            j = 0
        else:
            # next greater element
            q = rd[refi]
            md = np.inf
            for i in range(refi,len(rd)):
                if rd[i,axis] <= q[axis]:
                    continue
                if rd[i,axis] > q[axis]:
                    if rd[i,axis] - q[axis] < md:
                        md = rd[i,axis] - q[axis]
                        j = i
            if j == None:
                return None
        return self.break_point_tiebreakers(j,direction)

    def break_point_tiebreakers(self,refni,direction):
        #a = 1 if direction in {'l','r'} else 0
        if direction in {'l','r'}: 
            rd = self.xyl_sorted_y
            a = 1
        else:
            rd = self.xyl_sorted_x
            a = 0 

        rp = rd[refni]
        indices = np.where(rd[:,a] == rp[a])

        pp = rd[indices]
        s = None
        if direction == 'r':
            s = np.argmax(pp[:,0])
        elif direction == 'l':
            s = np.argmin(pp[:,0])
        elif direction == 't':
            s = np.argmax(pp[:,1])
        else:
            s = np.argmin(pp[:,1])

        ss = np.where((rd == pp[s]).all(axis=1))
        return ss[0][0]

    ######### start: delineation optimizer

    def optimize_delineation(self):
        s = 0 
        for (i,x) in enumerate(self.d.d_):
            s += self.improve_curve(x)
        return s

    def improve_curve(self,c):
        # get pertinent points to curve
        pp = self.pertinent_points_to_curve(c)
        # get curveset
        cs = self.d.complement_set_to_curve(c)

        # classify points
        s1 = self.classify_pertinent_points(c,cs,pp)
        # alternate and reclassify
        c.modulate_fit()
        s2 = self.classify_pertinent_points(c,cs,pp)

        if s1 > s2:
            c.modulate_fit()
            return 0 
        return s2 - s1

    def classify_pertinent_points(self,rc,cs,pp):

        s = 0
        for p in pp:
            l = self.d.classify_point_by_curveset(p[:2],rc,cs)
            if l != -1:
                s += 1
        return s

    def pertinent_points_to_curve(self,c):
        '''
        :return: np.ndarray, points pertinent to curve `c`
        '''
        
        ps = []
        for p in self.xyl:
            if c.in_point_range(p):
                ps.append(p)
        return ps

    ######### analyze delineation

    def analyze_delineation(self):
        '''
        method is called after `optimize_delineation`;
        delineation `d` is put into the cache, and
        delineation is done on the points in `d` for
        separability.

        :return: [indices of xyl points in delineation,\
                counter(label->count),score of delineation] 
        '''

        # iterate through all points and collect those in
        # delineation
        indices = []
        counter = Counter()
        q = 0 # number of correctly labelled points
        for (i,p) in enumerate(self.xyl):
            c = self.d.classify_point(p)
            if c != -1:
                indices.append(i)
                # case: classification equals label
                if c == int(p[2]): 
                    q += 1
                counter[int(p[2])] += 1
            else:
                if int(p[2]) == self.label:
                    counter[-1] += 1




            '''
                if c == int(p[2]): q += 1
            else:
                if int(p[2]) == self.label:
                    counter[-1] += 1
                counter[int(p[2])] += 1
            else:
                if p[2] 
                counter[c]
            '''

        return indices,counter,q

    def full_process(self,preprocess = True):
        '''
        main method
        '''
        if preprocess:
            self.preprocess()
            print("finished preprocessing ")
        self.collect_break_points()
        self.optimize_delineation()
        indices,c,q = self.analyze_delineation()
        # remove those indices from xyl
        self.xyl = np.delete(self.xyl,indices,0)
        return indices,c,q

def test_dataset__Dlineate22_1():
    '''
    x      x
    
    x   x       
    
    x   x      
                x
    x
    
    x      x
    '''
    return np.array([[5.,15.,0],\
        [5.,12.,0],\
        [5.,9.,0],\
        [5.,6.,0],\
        [5.,0.,0],\
        [15.,12.,0],\
        [15.,9.,0],\
        [20.,15.,0],\
        [20.,0.,0],\
        [25.,7.5,0]])

def test_dataset__Dlineate22_1_v2():
    d = test_dataset__Dlineate22_1()
    t2 = generate_random_xyl_points_at_center(\
        [12.,8.55,5],[[0.,5.,5],[15.,25.,8]],1)
    return np.vstack((d,t2))


def test_dataset__Dlineate22_1_v3(numPoints,rdistance):
    data = test_dataset__Dlineate22_1()
    dl = DLineate22(data)
    dl.preprocess()
    dl.collect_break_points()

    l = len(dl.d.d_)
    ps = 0
    points = []
    i = 0
    while ps < numPoints:
        a = random.random()
        if a > 0.5:
            pr = dl.d.d_[i].get_point()
            rp = random_point_near_2d_point_pair(\
                pr,rdistance)
            rp_ = [rp[0],rp[1],1.]
            points.append(rp_)
            ps += 1
        i = (i + 1) % l
    points = np.array(points)
    return np.vstack((data,points))

def test_dataset__Dlineate22_2():
    '''
    x      x

    
    x      x
    '''
    return np.array([[5.,15.,0],\
        [5.,0.,0],\
        [20.,15.,0],\
        [20.,0.,0]])


def test_dataset__Dlineate22_3():
    # NOTE: delineation does not work for triangles
    '''
       x   

    
    x      x
    '''
    return np.array([[5.,0.,0],\
        [7.5,15.,0],\
        [20.,0.,0]])


def test_dataset__Dlineate22_4():
    # NOTE: delineation does not work for triangles
    '''
       x   

    
    x      
            x
    '''
    return np.array([[5.,0.,0],\
        [7.5,15.,0],\
        [20.,-5.,0]])


def test_dataset__DLineateMC_1():
    '''
    completely separable; 4 labels.
    '''
    c1 = [0.,0.]
    drnp1 = [[0.,5.,1000],[5.,15.,10000],[15.,30.,300]]
    l1 = 0
    d1 = generate_random_xyl_points_at_center(c1,drnp1,l1)

    c2 = [80.,20.]
    drnp2 = [[0.,5.,500],[10.,20.,5000],[10.,15.,300]]
    l2 = 1
    d2 = generate_random_xyl_points_at_center(c2,drnp2,l2)

    c3 = [-50.,-30.]
    drnp3 = [[10.,15.,5000],[20.,25.,500],[25.,27.,300]]
    l3 = 2
    d3 = generate_random_xyl_points_at_center(c3,drnp3,l3)

    c4 = [-150.,-70.]
    drnp4 = [[10.,20.,5000],[20.,40.,5000]]
    l4 = 3
    d4 = generate_random_xyl_points_at_center(c4,drnp4,l4)
    
    d = d1
    d = np.vstack((d,d2))
    d = np.vstack((d,d3))
    d = np.vstack((d,d4))
    return d


from morebs2 import poly_factor,poly_struct
from copy import deepcopy
import unittest
import numpy as np

def pfe_case1():
    pa = np.array([[24,12],[-88,11],[176,10],[-286,9],[300,8],
            [-348,7],[465,6],[-252,5],[126,4],[-135,3]])
    p = poly_struct.CEPoly(pa)

    pfe1 = poly_factor.PolyFactorEst(p)
    pfe1.set_exponential_bounds((0,6),(3,6))
    return pfe1 

'''
python -m morebs2.tests.poly_factor_test  
'''
class TestPolyFactorEst(unittest.TestCase):

    def test__pfe__initial_bounds_hypothesis(self):
        pfe1 = pfe_case1()
        pfe1.initial_bounds_hypothesis()
        assert str(pfe1.p1) == ' 12x^6 + 45x^0'
        assert str(pfe1.p2) == ' 2x^6  -3x^3'

    def test__pfe__guess(self):
        pa = np.array([[24,12],[-88,11],[176,10],[-286,9],\
            [300,8],[-348,7],[465,6],[-252,5],[126,4],[-135,3]])

        p = poly_struct.CEPoly(pa)
        pfe1 = poly_factor.PolyFactorEst(deepcopy(p))
        pfe1.set_exponential_bounds((0,6),(3,6))
        pfe1.initial_bounds_hypothesis()
        pfe1.guess()
        assert pfe1.f1 * pfe1.f2 == p, "incorrect factorization"

    def test__pfe__guess__case2(self):
        pa = np.array([[1,2],[-1,0]])
        p = poly_struct.CEPoly(pa)
        pfe1 = poly_factor.PolyFactorEst(deepcopy(p))
        pfe1.set_exponential_bounds((0,1),(0,1))
        pfe1.initial_bounds_hypothesis()
        pfe1.guess()
        assert pfe1.f1 * pfe1.f2 == p, "incorrect factorization #2"

    def test__possible_new_elements_of_factor_for_exponent(self):
        pfe1 = pfe_case1()
        pfe1.initial_bounds_hypothesis()

        q = poly_factor.possible_new_elements_of_factor_for_exponent(\
            pfe1.p1,pfe1.p2,pfe1.p1eb,pfe1.p2eb,pfe1.p,6)
        assert q == [3, 2, 1]

        q = poly_factor.possible_new_elements_of_factor_for_exponent(\
            pfe1.p1,pfe1.p2,pfe1.p1eb,pfe1.p2eb,pfe1.p,8)
        assert q == [5, 4, 3, 2]

        q = poly_factor.possible_new_elements_of_factor_for_exponent(\
            pfe1.p1,pfe1.p2,pfe1.p1eb,pfe1.p2eb,pfe1.p,4)
        assert q == [1]

if __name__ == '__main__':
    unittest.main()


from morebs2 import point_sorter
import unittest
import numpy as np

'''
python -m morebs2.tests.point_sorter_test  
'''

def sample_pointsort_vector_1():
    return np.array([[0.0,5.0,12.0],\
        [0.0, 7.1, 6.3],\
        [1.0,15.0,31.0],\
        [2.0,10.0,1000.0],\
        [3.1,6.6,1.6],\
        [3.1,6.6,7.1],\
        [3.2,10.5,1.1],\
        [4.0,1.0,11.0],\
        [5.0,5.0,5.0],\
        [5.0,5.0,5.0]])

def test_tie_sets_for_vector():

    v1 = np.array((3.4, 3.4, 5.0, 5.01, 6.32, 6.32, 6.32, 6.32, 10.003))
    ts = point_sorter.tie_sets_for_vector(v1)

    assert [0,1] in ts, "missing tie 1"
    assert [4,7] in ts, "missing tie 2"
    assert len(ts) == 2, "incorrect ties #1"

    v2 = np.array([5.5,5.5])
    ts = point_sorter.tie_sets_for_vector(v2)
    assert [0,1] in ts and len(ts) == 1, "incorrect ties #2"
    return

class TestPointSorterClass(unittest.TestCase):

    def test__PointSorter_sort_it(self):

        q = sample_pointsort_vector_1()
        np.random.shuffle(q)
        ps = point_sorter.PointSorter(q)

        ps.sort_it()
        '''
        print("------")
        print("PS")
        print(ps.newData)
        '''
        testPointSortVector = sample_pointsort_vector_1()
        assert np.all(np.equal(ps.newData,testPointSortVector))


    def test_sorted_vector_value_search(self):

        tv1 = np.array([1.0, 2.0, 3.3, 5.2, 6.7, 8.9, 9.2, 11.3, 14.2])
        tv2 = np.array([6.7])
        tv3 = np.array([0.2, 0.4, 3.1, 6.7, 9.1, 20.1])
        tv4 = np.array([0.2, 0.4, 3.1, 6.71, 9.1, 20.1])

        assert point_sorter.sorted_vector_value_search(tv1, 6.7)[1]
        assert point_sorter.sorted_vector_value_search(tv2, 6.7)[1]
        assert point_sorter.sorted_vector_value_search(tv3, 6.7)[1]
        assert not point_sorter.sorted_vector_value_search(tv4, 6.7)[1]

    def test_sorted_vector_value_range_search(self):
        q = sample_pointsort_vector_1()
        vec = q[:,0]
        vec = np.sort(list(vec) + [3.1])
        q = point_sorter.sorted_vector_value_range_search(vec, 3.1)
        assert q[0] == [4,6], "invalid range search #1"

        vec2 = np.array([6.6,6.6])
        q = point_sorter.sorted_vector_value_range_search(vec2, 6.6)
        assert q[0] == [0,1], "invalid range search #2"

        vec3 = np.array([0.2, 0.4, 0.4, 0.4, 3.1, 6.71, 9.1, 20.1])
        q = point_sorter.sorted_vector_value_range_search(vec3, 0.4)
        assert q[0] == [1,3]

        vec4 = np.array([0.2, 0.4, 3.1, 6.71, 9.1, 20.1])
        q = point_sorter.sorted_vector_value_range_search(vec4, 0.4)
        assert q[0] == [1,1]

    def test__PointSorter__vector_exists(self):

        q = sample_pointsort_vector_1()
        np.random.shuffle(q)
        ps = point_sorter.PointSorter(q)
        ps.sort_it()

        v = np.array([3.1,6.6,7.10])
        assert ps.vector_exists(v) != -1

        v2 = np.array([3.1,6.6,7.12])
        assert ps.vector_exists(v2) == -1

        v3 = np.array([5.0,5.0,5.0])
        assert ps.vector_exists(v3) != -1

    def test__PointSorter__insert_point(self):
        q = sample_pointsort_vector_1()
        np.random.shuffle(q)
        q2 = np.empty((0,q.shape[1]))
        ps = point_sorter.PointSorter(q2)

        for q_ in q:
            ps.insert_point(q_)

        q = sample_pointsort_vector_1()
        assert np.all(np.equal(ps.newData,q))

    def test__PointSorter__update_points_from_cache(self):
        q = sample_pointsort_vector_1()
        np.random.shuffle(q)
        q2 = np.empty((0,q.shape[1]))
        ps = point_sorter.PointSorter(q2)

        for q_ in q:
            ps.log_delta_cache(q_,1)

        ps.update_points_from_cache(np.array([5.0,5.0,5.0]))

        assert point_sorter.equal_iterables(ps.extremumData[-1][0],np.array([5.0,5.0,5.0]))
        assert point_sorter.equal_iterables(ps.extremumData[1][0],[   2.,   10., 1000.])

        # delete extremum 1 and recheck `extremumData`
        x = [   2.,   10., 1000.]
        ps.log_delta_cache(x,-1)

        ps.update_points_from_cache(np.array([5.0,5.0,5.0]))

        assert point_sorter.equal_iterables(ps.extremumData[-1][0],np.array([5.0,5.0,5.0]))
        assert point_sorter.equal_iterables(ps.extremumData[1][0],[ 1., 15., 31.])

if __name__ == "__main__":
    unittest.main()



from morebs2 import chained_poly_interpolation
import numpy as np
import unittest

'''
python -m morebs2.tests.chained_poly_interpolation_test  
'''
class TestChainedLagrangePolySolverClass(unittest.TestCase):

    """
    1 y
    2 n
    """
    def LagrangePolySolver___duplicates_exist(self):
        # duplicates work 1
        x1 = [1.0,4.2]
        x2 = [5.0,4.2]
        x3 = [5.0,4.2]
        #ps = np.array([x1,x2])
        ps = np.array([x1,x2,x3])
        lps = chained_poly_interpolation.LagrangePolySolver(ps)
        ps2 = lps.form_point_sequence()

        # duplicates work 2
        x1 = [1.0,4.2]
        x2 = [5.0,4.2]
        x3 = [5.0,4.3]
        ps = np.array([x1,x2,x3])
        lps = chained_poly_interpolation.LagrangePolySolver(ps)
        ps2 = lps.form_point_sequence()

        ##
        """
        print("TESTING")
        for p in ps2:
            print(p)

        print("CASE 3")
        r = lps.output_by_lagrange_basis(1.0)
        print(r)
        """

    def test__ChainedLagrangePolySolver__init___failed(self):

        # a 4-d case 1
        x1 = [1.0,4.2,7.3,1.0]
        x2 = [5.0,4.2,7.3,2.0]
        x3 = [7.0,4.2,7.3,3.0]
        xs = np.array([x1,x2,x3])
        try:
            clps = chained_poly_interpolation.ChainedLagrangePolySolver(xs)
            assert False, "failioncos"
        except:
            #print("failed point-set 1.1")
            pass

        # a 4-d case 2
        x1 = [1.0,4.2,7.3,1.0]
        x2 = [5.0,4.5,7.2,2.0]
        x3 = [7.0,3.0,7.1,3.0]
        x4 = [1.0,5.0,7.0,1.0]
        x5 = [5.0,4.2,7.05,2.0]
        x6 = [7.0,4.2,6.8,3.0]
        xs = np.array([x1,x2,x3,x4,x5,x6])

        clps = chained_poly_interpolation.ChainedLagrangePolySolver(xs)#.form_point_sequence(hopIncrement = hopIncrement,capture = False)
        clps.set_axis_order_of_interpolation(np.array([3,1,0,2]))
        assert not clps.check_valid_ordering(), "invalid case 2"
        return

    def test__ChainedLagrangePolySolver__at_x1(self):

        # a 4-d case 2
        x1 = [1.0,4.2,7.3,1.0]
        x2 = [15.0,4.5,7.2,5.0]
        x3 = [7.5,1.0,1.1,3.0]
        x4 = [3.0,2.3,2.7,4.0]
        x5 = [2.0,0.29,3.05,12.0]
        x6 = [10.0,3.2,4.8,3.3]
        xs = np.array([x1,x2,x3,x4,x5,x6])

        clps = chained_poly_interpolation.ChainedLagrangePolySolver(xs)
        clps.set_axis_order_of_interpolation(np.array([0,1,2,3]))
        clps.formulate()

        q = clps.points[:,0]
        for (i,q_) in enumerate(q):
            p = clps.at_x1(q_)
            """
            print("point")
            print(p)
            print("actual")
            print(clps.points[i])
            """
            assert chained_poly_interpolation.equal_iterables(p,clps.points[i])
        return

if __name__ == '__main__':
    unittest.main()



from .numerical_generator import *
from .point_weight_function import *

#################### start: area estimation methods

"""
table contains values of curvature estimation between two
circles that intersect.

These values are used to `gauge` the volume of intersection between
two n-balls.
"""
CURVATURE_EST = {1.0:0.46792,\
                5/6:0.52544,\
                2/3:0.61026,\
                1/2:0.72842,\
                1/3:0.88860,\
                1/4:0.96300,\
                1/6:1.0}

CURVATURE_EST_KEYS = [1/6,1/4,1/3,1/2,2/3,5/6,1]

def intersection_ratio(b1,b2):
    """
Calculates the intersection ratios of b1 and b2.
The intersection ratio of a ball is the proportion of its diameter that
lies in the intersection.

:return: intersection ratio of b1, intersection ratio of b2
:rtype: (float,float)|(None,None)
    """
    ed = euclidean_point_distance(b1.center,b2.center)

    # case: no intersection
    if ed >= b1.radius + b2.radius:
        return None,None

    # case: b2 completely in b1
    if b2.radius + ed <= b1.radius:
        return 0.0,1.0

    # case: b1 completely in b2
    if b1.radius + ed <= b2.radius:
        return 1.0,0.0

    # get intersection ratio of b1
        # first point is the farthest boundary point of b1 that lies in b2
        # point on ray (b1.center,b2.center) of distance b1.radius from .b1.center
    coeff1 = b1.radius / ed
    x1 = b1.center + coeff1 * (b2.center - b1.center)

        # second point of interest is the farthest boundary point of b2 that lies in b1
        # the point on the ray (b2.center,b1.center) at
        # distance b1.radius + ed
    coeff2 = b2.radius / ed
    x2 = b2.center + coeff2 * (b1.center - b2.center)
    distance = euclidean_point_distance(x1,x2)
    return round(distance / (b1.radius * 2),5), round(distance / (b2.radius * 2),5)

def prelim_2intersection_estimate(b1,ir):
    """
    estimate does not include curvature estimation
    """
    if ir <= 0.5:
        return ball_area(b1.radius, b1.center.shape[0]) * ir
    return ball_area(b1.radius,b1.center.shape[0]) * 0.5 + prelim_2intersection_estimate(b1,ir - 0.5)

def reference_intersection_ratio_to_curvature_est_key(ri):
    assert ri >= 0 and ri <= 1, "invalid intersection ratio"
    x = np.array([abs(ri - k) for k in CURVATURE_EST_KEYS])
    i = np.argmin(x)
    return CURVATURE_EST_KEYS[i]

# CAUTION
##this is an approximation for Ball volume.
##Value calculated from .this method is to be used as a reference
##rather than directly as a numerical value.
def ball_area(r,k):
    assert k >= 2,"invalid k"
    x = math.pi * float(r) ** 2
    x2 = (k - 1) ** 2
    return x * x2

def reference_volume_ratio(b1,b2):
    return min([b1.radius / b2.radius,b2.radius / b1.radius])

def volume_2intersection_estimate(b1,b2):
    r1,r2 = intersection_ratio(b1,b2)
    ###print("RI: ",r1,r2, ball_area(b1.radius,6),ball_area(b2.radius,6))
    # case: no intersection
    if type(r1) == type(None):
        return 0.0

    # case: ball 1 in ball 2
    if r1 == 1.0:
        return ball_area(b1.radius,b1.center.shape[0])

    # case: ball 2 in ball 1
    if r2 == 1.0:
        return ball_area(b2.radius,b2.center.shape[0])

    # case: intersection
    ie1 = prelim_2intersection_estimate(b1,r1)
    ie2 = prelim_2intersection_estimate(b2,r2)

    ###print("IE1 ", ie1, "\t",ie2)

    referenceIntersectionRatio = r1 if b1.radius > b2.radius else r2
    cek = reference_intersection_ratio_to_curvature_est_key(referenceIntersectionRatio)
    rv = reference_volume_ratio(b1,b2)
    curvatureEst = CURVATURE_EST[cek] + (1.0 - CURVATURE_EST[cek]) * rv
    return min([ie1,ie2]) * curvatureEst

#################### end: area estimation methods

class Ball:
    '''
    implementation of n-ball, see `ball_operator` for related methods.

    :param center: center of ball
    :type center: np.ndarray, vector
    :param idn: int
    :type idn: int
    '''
    # used for reversion
    DELTA_MEMORY_CAPACITY = 3

    def __init__(self, center,idn = None):
        assert is_vector(center), "invalid center point for Ball"
        self.center = center
        self.idn = idn
        self.data = PointSorter(np.empty((0,self.center.shape[0])))
        self.radius = 0.0

        # [0] is radial reference, [1] is
        self.radiusDelta = (None,None)
        self.radiusDeltas = []
        self.pointAddDeltas = []

        # container that holds the most recent delta
        self.clear_delta() # (point,radius delta)
        self.neighbors = set() # of ball neighbor identifiers
        return

    def __str__(self):
        s0 = "ball idn: {}".format(self.idn)
        s = "\nball center:\n\t{}".format(vector_to_string(self.center,float))
        s2 = "\nball radius: {}".format(self.radius)
        s3 = "\nball neighbors: {}".format(self.neighbors)
        s3 = "\nball data shape: {}".format(self.data.newData.shape)

        return s0 + s + s2 + s3
    '''
    '''
    @staticmethod
    def dataless_copy(b):
        b_ = Ball(np.copy(b.center))
        b_.radius = b.radius
        b_.neighbors = set(b.neighbors)
        return b_

    def is_neighbor(self,b):
        return euclidean_point_distance(self.center,b.center) <= self.radius + b.radius

    def area(self):
        b = ball_area(self.radius, self.center.shape[0])
        return round(b,5)

    @staticmethod
    def one_ball(center, points):
        b = Ball(center)

        for p in points:
            b.add_element(p)
        return b

    @staticmethod
    def one_ball_(center,radius,centerAsPoint = True):
        b = Ball(center)
        c = random_npoint_from_point(np.copy(center),radius)
        b.add_element(c)
        b.add_element(center)
        return b

    # TODO: test this.
    def in_bounds(self,bounds):
        assert self.center.shape[0] == bounds.shape[0], "invalid bounds"

        bs = []
        for (i,c) in enumerate(self.center):
            c0,c1 = c - self.radius,c + self.radius
            bs.append((c0,c1))
        bs = np.array(bs)
        bs = np.round(bs,5)
        return bounds_is_subbounds(bounds,bs)

    def point_in_data(self,p):
        return self.data.vector_exists(p)

    def point_in_ball(self,p):
        ed = euclidean_point_distance(p,self.center)
        return ed <= self.radius

    '''
    adds single point to Ball
    and updates mean
    '''
    def add_element(self, p):
        self.data.insert_point(p)
        self.pointAddDeltas.insert(0,p)
        self.pointAddDeltas = self.pointAddDeltas[:Ball.DELTA_MEMORY_CAPACITY]

        # update radius
        r = euclidean_point_distance(self.center,p)
        if r > self.radius:
            self.radiusDeltas.insert(0,self.radiusDelta)
            self.radiusDeltas = self.radiusDeltas[:Ball.DELTA_MEMORY_CAPACITY]
            self.radiusDelta = (p, r - self.radius)
            self.radius = r

    def revert_add_point(self):
        if len(self.pointAddDeltas) == 0: return
        q = self.pointAddDeltas.pop(0)

        # remove from .data
        if type(self.radiusDelta[0]) != type(None):
            if equal_iterables(q,self.radiusDelta[0]):
                self.revert_delta()
        else:
            self.data.delete_point(q)
        return

    # TODO:
    '''
    '''
    def revert_delta(self):
        if type(self.radiusDelta[0]) == type(None):
            return
        self.data.delete_point(self.radiusDelta[0])
        self.radius = self.radius - self.radiusDelta[1]

        self.radiusDelta = self.radiusDeltas.pop(0) if len(self.radiusDeltas) > 0\
                        else (None,None)

        return

    def clear_delta(self):
        self.radiusDelta = (None,None)

    '''
    determines if two balls intersect
    '''
    @staticmethod
    def does_intersect(b1,b2):
        epd = euclidean_point_distance(b1.center,b2.center)
        epd = epd - b1.radius - b2.radius
        return epd <= 0

    '''
    upper-bound estimate on the k-dimensional sub-area of
    intersection b/t b1 and b2.
    '''
    @staticmethod
    def area_intersection_estimation(b1,b2):
        a1 = Ball.area_intersection_estimation_(b1,b2)
        a2 = Ball.area_intersection_estimation_(b2,b1)
        x1 = min([b2.area() * a1, b1.area() * a2])
        return max([0.0,x1])


    '''
    merges two balls
    '''
    def __add__(self, b2):

        r1,r2 = intersection_ratio(self,b2)
        # case: self in b2
        if r1 == 1.0:
            return b2

        # case: b2 in self
        if r2 == 1.0:
            return self

        # case: other
        ed = euclidean_point_distance(self.center,b2.center)
            # get the farthest boundary point of b2 on line (b1.center,b2.center)
        x2 = b2.radius / ed
        e2 = b2.center + x2 * (b2.center - self.center)

            # get the farthest boundary point of b1 on line (b1.center,b2.center)
        x1 = self.radius / ed
        e1 = self.center + x1 * (self.center - b2.center)

            # get midpoint of line
        center = np.round((e1 + e2) / 2.0,5)
        b3 = Ball(center)
        b3.radius = np.round(euclidean_point_distance(e1,e2) / 2.0,5)
        dr = np.vstack((self.data.newData,b2.data.newData))
        b3.data = PointSorter(dr)
        return b3



from morebs2 import hop_pattern
import unittest

# TODO: check reversing bounds [1,0] <-> [0,1]
'''
python -m morebs2.tests.hop_pattern_test  
'''
class TestHopPatternClass(unittest.TestCase):

    def test__HopPattern__calculate_hop_directions(self):

        # tests hop_one method by calling it
        # for a cycle

        # case 0:
        q = hop_pattern.HopPattern(23.0,23.0,97.0)
        q.calculate_hop_directions()
        assert q.hopDirection.shape == (2,)
        assert abs(q.hopDirection[0] - (97 - 23.00) / 5) <= 10 ** -5
        ##return -1

        # case 1:
        q = hop_pattern.HopPattern(97.0,23.0,97.0)
        q.calculate_hop_directions()
        assert abs(q.hopDirection[0] + (97 - 23.00) / 5) <= 10 ** -5, "incorrect hop directions {}".format(q.hopDirection)
        assert q.hopDirection[0] < 0.0, "incorrect hop direction"

        # case 2:
        q = hop_pattern.HopPattern(50.0,23.0,97.0)
        q.calculate_hop_directions()
        assert abs(q.hopDirection[0] - (97 - 23.00) / 5) <= 10 ** -5


    def test__HopPattern__hop_one(self):

        q = hop_pattern.HopPattern(50.0,23.0,97.0, DIR = 0.2)
        q.calculate_hop_directions()

        assert not q.did_cycle(), "invalid cycle register"

        c = 0
        for i in range(30):
            next(q)
            c += 1

            if (c - 1) % 5 == 0 and c > 1:
                assert q.did_cycle()
            else:
                assert not q.did_cycle()
        return

    def test__HopPattern__hop_one_2(self):
        hop_pattern.HopPattern.DEF_INCREMENT_RATIO = 0.2
        q = hop_pattern.HopPattern(23.0,23.0,97.0)
        q.calculate_hop_directions()

        c = 0
        while not q.did_cycle():
            q2 = next(q)
            c += 1
        assert c == 6, "cycle is incorrect length"

    """
    Runs HopPattern on the following range:
    [0,1]
    """
    def test__HopPattern__hop_one_4(self):
        q = hop_pattern.HopPattern(1.0,0.0,1.0, DIR = 0.5)
        q.calculate_hop_directions()
        c = 0
        x = [1.0, 0.5]

        for i in range(10):
            assert next(q) == x[c % 2], "incorrect "
            c += 1
        return

    def test__HopPattern__hop_one_5(self):
        q = hop_pattern.HopPattern(0.0,0.0,1.0, DIR = round(1/3,10))
        q.calculate_hop_directions()
        next(q)
        c = 1
        for i in range(10):
            #print(next(q))
            next(q)
            c += 1

            if c % 3 == 0:
                assert q.did_cycle()
            else:
                assert not q.did_cycle()

            c += 1

    ### HERE: do some tests on non-endpoint startpoint,
    ###       transfer those tests over to
    def test__HopPattern__hop_one_6(self):
        q = hop_pattern.HopPattern(50.0, 23.0, 97.0, DIR = round(1/3,10))
        q.calculate_hop_directions()

        c = 0
        next(q)
        while c < 100:
            next(q)
            c += 1
            if c % 3 == 0: assert q.did_cycle(), "q is supposed to cycle."
        return

    """
    """
    def test__HopPattern__modulo_hop(self):

        assert hop_pattern.HopPattern.modulo_hop(0.0, 0.0, [0.0,1.0], 0) == 0.0
        assert hop_pattern.HopPattern.modulo_hop(0.0, 0.0, [0.0,1.0], 1) == 1.0

        assert hop_pattern.HopPattern.modulo_hop(0.0, -0.5, [0.0,1.0], 0) == 0.5

        assert hop_pattern.HopPattern.modulo_hop(0.5, 0.5, [0.0,1.0], 1) == 1.0
        assert hop_pattern.HopPattern.modulo_hop(0.5, 0.5, [0.0,1.0], 0) == 0.0

    # TODO: add more cases
    def test__HopPattern__cycle_check(self):

        # case 1
        q = hop_pattern.HopPattern(1.0,0.0,1.0, DIR = 0.5)
        q.calculate_hop_directions()

        q2 = next(q)
        assert q2 == 1.0 and not q.cycle_check()

        q2 = next(q)
        assert q2 == 0.5 and not q.cycle_check()

        q2 = next(q)
        assert q2 == 1.0 and q.cycle_check()

        # case 2
        q = hop_pattern.HopPattern(0.0,0.0,1.0, DIR = 0.5)
        q.calculate_hop_directions()

        q2 = next(q)
        assert q2 == 0.0 and not q.cycle_check()

        q2 = next(q)
        assert q2 == 0.5 and not q.cycle_check()

        q2 = next(q)
        assert q2 == 0.0 and q.cycle_check()

    def test__HopPattern__rev__next__(self):
        # case 1
        q = hop_pattern.HopPattern(0.0,0.0,1.0, DIR = 0.5)
        x = [0.5, 1.0]
        next(q)

        for i in range(10):
            q2 = q.rev__next__()
            i = i % 2
            assert q2 == x[i]

        # case 2
        q = hop_pattern.HopPattern(1.0,0.0,1.0, DIR = 0.5)
        q.calculate_hop_directions()
        x = [0.5, 0.0]
        next(q)
        for i in range(10):
            q2 = q.rev__next__()
            i = i % 2
            assert q2 == x[i]

        # case 3
        q = hop_pattern.HopPattern(1.0,0.0,1.0, DIR = round(1/3,10))
        q.calculate_hop_directions()
        x = [1.0, round(1/3,5), round(1/3,5) * 2]
        for i in range(10):
            q2 = q.rev__next__()
            i = i % 3
            assert q2 - x[i] < 10 ** -5, "wrong"
        return

def test__HopPattern__rev__next__2():
    q = HopPattern(0.5,0.0,1.0,DIR = round(1/3,10))
    # get hop info
    q.rev__next__()
    q.rev__next__()
    assert not q.did_cycle()
    q.rev__next__()
    q.rev__next__()
    assert q.did_cycle()

def test__HopPattern__rev__next__3():
    q = HopPattern(0.5,1.0,0.0,DIR = round(1/3,10))
    # get hop info
    for i in range(10):
        print("Q ", q.rev__next__())

    return -1

    q.rev__next__()
    q.rev__next__()
    assert not q.did_cycle()
    q.rev__next__()
    q.rev__next__()
    assert q.did_cycle()

if __name__ == '__main__':
    unittest.main()



from morebs2 import rssi,relevance_functions
#from relevance_functions import * 

import unittest
import numpy as np

'''
python -m morebs2.tests.relevance_functions2_test  
'''
class TestRelevanceFunctions(unittest.TestCase):

    def test__sample_rch_1_with_update(self):
        b = np.array([[0,1.0],[0,1.0],[0,1.0]])
        pb = np.copy(b)
        h = 9
        cv = 0.4
        rch = relevance_functions.sample_rch_1_with_update(b,pb,h,cv)

        # do one
        p = np.array([0.5,0.5,0.5])
        q = rch.apply(p)
        assert not q, "failed case 1"

        # update args
        pb2 = np.array([[0,0.6],[0,0.6],[0,0.6]])
        pb2 = np.array([[0,0.5],[0,0.5],[0,0.5]])

        updateArgs = [b,pb2,h,cv]
        rch.load_update_vars(updateArgs)
        rch.update_rch()

        p2 = np.array([0.3,0.3,0.3])
        p2 = np.array([0.25,0.25,0.25])

        q = rch.apply(p2)
        assert not q, "failed case 2"


    def sample_rssi_1_with_update(self):
        """
        an RSSI instance with an updating RCH. RSSI runs in mode::(relevance zoom)
        """
        b = np.array([[0,1.0],[0,1.0],[0,1.0]])
        pb = np.copy(b)
        h = 9
        cv = 0.4

        rch = relevance_functions.sample_rch_1_with_update(b,pb,h,cv)
        ##
        r = ResplattingSearchSpaceIterator(b, np.copy(b[:,0]), None, h,\
            resplattingMode = ("relevance zoom",rch), additionalUpdateArgs = (cv,))
        return r


if __name__ == '__main__':
    unittest.main()



'''
(a(X_n) + c) modulo m
'''
import random
from math import ceil
from .matrix_methods import *
from .variance_works import *
from collections import OrderedDict

def modulo_in_range(i,r): 
    assert r[0] < r[1], "invalid range {}".format(r) 
    return i % (r[1] - r[0]) + r[0] 

class LCG:
    '''
    linear congruential generator
    '''
    
    def __init__(self, startingInteger, modulo):
        assert type(modulo) is int and modulo > 128, "invalid modulo"
        assert type(startingInteger) is int and startingInteger < modulo, "invalid starting integer"

        self.startingInteger = startingInteger
        self.vInt = startingInteger
        self.modulo = modulo
        self.fetch_arguments()

    # TODO: mod here?
    def fetch_arguments(self):
        self.multiplier = self.fetch_random_integer_in_range((0, self.modulo))
        self.increment = self.fetch_random_integer_in_range((0, self.modulo))

    def fetch_random_integer_in_range(self, ranje):
        assert type(ranje[0]) is int and type(ranje[1]) == type(ranje[0]), "invalid range"
        assert ranje[0] < ranje[1], "invalid range [1]"

        q = random.random()
        c = ceil(ranje[0] + q * (ranje[1] - ranje[0]))
        return int(c)

    def calculate(self, v):
        assert type(v) in {int,float,np.int32,np.int64,\
            np.float32,np.float64}, "invalid value v,type {}".format(type(v))
        return (self.multiplier * v + self.increment) % self.modulo

    def __next__(self):
        q = self.vInt
        self.vInt = self.calculate(q)
        return q


def prg__LCG(start,multiplier,increment,modulo):

    lcg = LCG(1,np.iinfo(np.int32).max) 
    lcg.startingInteger = start 
    lcg.vInt = start 
    lcg.modulo = modulo
    lcg.multiplier = multiplier
    lcg.increment = increment 
    return lcg.__next__ 

#############------------------------------------------------------------

# NOTE: __next__ not working. 
class CycleMap:

    '''
    '''
    def __init__(self, cRange:'int'):
        assert type(cRange) is int, "invalid type for cycle range"
        self.cRange = cRange
        self.mahp = OrderedDict()
        self.c = None 
        self.head = None
        return

    def set_map(self, m):
        assert type(m) is OrderedDict, "invalid type for map"
        assert CycleMap.is_valid_map(m), "invalid cycle map"
        assert len(m) == self.cRange, "invalid length for map"
        self.mahp = m

    @staticmethod
    def random_cycle_map(vortexRange):
        ks = np.arange(vortexRange)
        np.random.shuffle(ks)

        # select the head
        h = ks[0]
        h_ = h
        l = OrderedDict()
        rem = np.arange(vortexRange)
        for i in range(vortexRange - 1):
            possible = [r for r in rem if r != h]

            # choose random
            lp = len(possible)
            ri = random.randrange(lp)
            l[h] = possible[ri]

            rem = np.array(possible)
            h = l[h]

        l[h] = h_

        return l

    @staticmethod
    def is_valid_map(m):
        if len(m) == 0: 
            return False 

        # fetch the first element in the map
        q = list(m.keys())[0]

        # check for cycle
        x = [q]
        l = len(m)
        c = 0
        while c < l:
            r = m[x[-1]]
            x.append(r)
            c += 1

        # check that last element is first key
        if not (x[-1] == x[0]): return False
        # check that number of unique elements is l
        if len(set(x)) != l: return False
        return True

    def head_(self):
        for k,v in self.mahp.items():
            self.head = k
            self.c = k
            break

    def __next__(self):
        if type(self.c) == type(None): self.head_()
        q = self.mahp[self.c]
        self.c = q
        return q

    def v(self,k):
        return self.mahp[k]

    #### TODO: make generators for non-cycles
    
#------------------------------------------------------------------
#### binary sequence generators

def generate_possible_binary_sequences(vecOrder, thisNumber, elements = [0,1]):

    if len(thisNumber) == vecOrder:
        yield thisNumber
        return

    q1, q2 = np.copy(thisNumber), np.copy(thisNumber)
    q1, q2 = np.hstack((q1,[elements[0]])), np.hstack((q1,[elements[1]]))
    yield from generate_possible_binary_sequences(vecOrder, q1,elements)
    yield from generate_possible_binary_sequences(vecOrder, q2,elements)

def generate_random_binary_sequence(vecOrder):
    assert type(vecOrder) is int, "invalid vec. order"
    return np.random.randint(0, 2, (vecOrder,))

def prg__constant(x=0): 

    def f(): 
        return x
    return f 

class ModuloAlternator: 

    def __init__(self,s0,s1,s=None): 
        assert s1 > s0 
        assert type(s1) == type(s0)
        assert type(s1) == int 

        self.s0 = s0 
        self.s1 = s1 
        if s == None: 
            self.s = self.s0 
        else: 
            self.s = (s % self.s1) + self.s0 

    def __next__(self): 
        q = self.s
        self.s += 1
        if self.s >= self.s1: 
            self.s = (self.s % self.s1) + self.s0 
        return q 


def prg__n_ary_alternator(s0=0,s1=2,start=0): 
    ma = ModuloAlternator(s0,s1,start) 
    return ma.__next__ 

####---------------------------------------------------------------------
#### uniform dist. numerical generators

"""
"""
def generate_uniform_sequence_in_bounds(vecOrder, bounds,rnd_struct=rng):
    assert is_2dmatrix(bounds), "invalid bounds {}".format(bounds)
    assert vecOrder == len(bounds) or len(bounds) == 1, "invalid bounds"

    if len(bounds) == 1:
        return rnd_struct.uniform(bounds[0,0], bounds[0,1], (vecOrder,))
    else:
        q = np.zeros((vecOrder,))
        for i in range(vecOrder):
            q[i] = rnd_struct.uniform(bounds[i,0], bounds[i,1])#, (vecOrder,))
        return q

def random_bounds_edge(bounds):
    assert is_bounds_vector(bounds), "invalid bounds"
    bs = generate_random_binary_sequence(bounds.shape[0])
    r = [i for i in range(bounds.shape[0])]
    return bounds[r,bs]

################################# TODO: noise methods need to be checked.


def k_random_points_in_bounds(minVec,maxVec,k):
    '''
    A method that generates `k` points in bounds [`minVec`,`maxVec`] by random.random().

    :param minVec: n-dimensional vector with each dimension the minumum value
    :type minVec: np.ndarray
    :param maxVec: n-dimensional vector with each dimension the maximal value
    :type maxVec: np.ndarray
    :param k: number of points to generate
    :type k: int, > 0
    '''
    assert np.all(maxVec - minVec >= 0.0), "invalid arguments min.,max. vec"

    d = maxVec - minVec
    for i in range(k):
        x = random.random()
        yield minVec + (d * x)

###------------------------

def one_random_noise(bounds,noiseRange):
    """
    :param bounds: bounds matrix
    :type bounds: np.ndarray, n x 2 matrix
    :param noiseRange: 2dmatrix, len is 1 or bounds.shape[0]
    """
    assert is_proper_bounds_vector(bounds), "invalid bounds"

    # set max distance for each dim.
    q = bounds[:,1] - bounds[:,0]
    us = generate_uniform_sequence_in_bounds(bounds.shape[0], noiseRange)
    q = q * us
    return q

def random_npoint_from_point(p1,l,roundDepth=5):
    p2 = np.copy(p1)
    r = np.arange(p1.shape[0])
    np.random.shuffle(r)
    r,x = r[:-1],r[-1]
    d = 0.0

    def delta_extreme():
        return math.sqrt(l ** 2 - d)

    def random_delta(j,randomize = True):
        delta = delta_extreme()
        delta = delta if random.random() > 0.5 else -delta

        if randomize:
            delta = random.uniform(0.0,delta)
        return delta

    for r_ in r:
        rd = random_delta(r_)
        p2[r_] = p2[r_] + rd
        d += (rd **2)
    p2[x] = p2[x] + random_delta(x,False)
    return np.round(p2,5)

def random_npoint_from_point_in_bounds_(b,p1,l,roundDepth = 5):
    """
    calculates a random n-dimensional point p2 in bounds `b` of distance `l`
    from `p1`.
    """

    p2 = np.copy(p1)
    r = np.arange(p1.shape[0])
    np.random.shuffle(r)
    r,x = r[:-1],r[-1]
    d = 0.0

    def delta_extreme():
        return math.sqrt(l ** 2 - d)

    def possible_extreme(j):
        return [b[j,0] - p1[j],b[j,1] - p1[j]]

    def random_delta(j,randomize = True):
        delta = delta_extreme()
        delta = delta if random.random() > 0.5 else -delta

        if randomize:
            delta = random.uniform(0.0,delta)

        pe = possible_extreme(j)
        pe = pe[1] if delta > 0 else pe[0]
        sols = [pe,delta]
        k = np.argmin(np.abs(sols))
        return sols[k]

    for r_ in r:
        rd = random_delta(r_)
        p2[r_] = p2[r_] + rd
        d += (rd **2)
    p2[x] = p2[x] + random_delta(x,False)
    return np.round(p2,5)

def random_npoint_from_point_in_bounds(b,p1,l,attempts = 15):
    t = False

    while not t and attempts > 0:
        p2 = random_npoint_from_point_in_bounds_(b,p1,l)
        #print("DISTANCE ", euclidean_point_distance(p1,p2))
        t = abs(euclidean_point_distance(p1,p2) - l) < 10 ** -5
        attempts -= 1
    return p2 if t else None

def add_noise_to_points_restricted_bounds(minVec,maxVec, points, noiseRange = np.array([[0.25,0.65]]), boundsRestriction = False):
    '''
    adds noise to points in restricted bounds by noise range,
    set to (0.25,0.65) by default.

    each output in yield is a point that may not lie within the bounds set by minVec
    and maxVec
    '''

    b = np.array([minVec,maxVec]).T
    for p in points:
        yield p + one_random_noise(b,noiseRange)

def random_noise_sequence(s,b,noiseRange):
    rn = []
    for i in range(s):
        n = one_random_noise(b,noiseRange)
        rn.append(n)
    return np.array(rn)

# TODO:
def generate_gaussian_sequence_in_bounds(mean, var):
    rng.normal()
    return -1

#-------------------------- sequence sorters using prgs

def prg_seqsort(l,prg): 
    """
    sorts a sequence using a pseudo-random number generator 

    :param l: sequence of elements; all elements of same type. 
    :type l: list 
    :param prg: pseudo-random number generator used to choose indices. 
    :type prg: function, no parameters. 
    """
    l_ = [] 
    while len(l) > 0: 
        i = prg() % len(l) 
        l_.append(l.pop(i))
    return l_ 

# TODO: unit-test this more 
def prg_seqsort_ties(l,prg,vf): 
    """
    sorts a sequence l

    :param l: sequence of elements; all elements of same type. 
    :type l: list 
    :param prg: pseudo-random number generator used to choose indices. 
    :type prg: function, no parameters. 
    :param vf: value-access function for element of l
    :type vf: function, one parameter is instance of element type for `l`. 
    """

    # sort l by vf first 
    q = sorted(l,key=vf)
    Q = [] 
    while len(q) > 0: 
        x = q.pop(0) 
        y = vf(x) 
        q2 = [x]

        # collect all ties with x
        j = -1 
        for (i,q_) in enumerate(q): 
            y2 = vf(q_) 
            if y == y2: 
                j = i 
            else: 
                break 
        j += 1 
        q2.extend(q[:j])
        while j > 0: 
            q.pop(0) 
            j -= 1 
        # permute q2 by prg 
        q2 = prg_seqsort(q2,prg)
        Q.extend(q2) 
    return Q 


def prg_partition_for_sz(S,num_sets,prg,variance):  
    """
    outputs a partition P (list) with positive integers 
    of length `num_sets` that sums to S. Uses the argument 
    `prg` that acts as a pseudo-random number generator 
    to output values with respect to `variance`. 

    NOTE: output may appear wrong according to `variance`. 
          The partitioning scheme assigns 1 to every set  
          in the beginning. Then it iterates through the 
          sets, P, and draws a number using the `prg` that is 
          at most the remaining number left, S - sum(P). This 
          drawing is what the `variance` measure relates to. 
          The range of possible numbers decreases the further 
          down the iteration. 
    """
    
    assert variance >= 0.0 and variance <= 1.0 
    assert S >= num_sets 

    P = [1 for _ in range(num_sets)] 
    S -= sum(P)

    i = 0 
    stat = S > 0 and i < num_sets 
    while stat:
        if i + 1 == num_sets: 
            P[-1] += S 
            break
        if S < 0: break  

        x = ceil(S / num_sets) 
        rem = ceil((S - x) * variance)

        rng = [x - rem, x + rem]

        if rng[0] == rng[1]: 
            P[i] += rng[0] 
            S -= rng[0] 
        else: 
            q = modulo_in_range(prg(),rng)
            q = modulo_in_range(prg(),[1,S+1]) 
            P[i] += q
            S -= q
        i += 1 

    return P 

def prg_choose_n(Q,n,prg,is_unique_picker:bool=False):
    l = len(Q)
    assert l > 0 

    if is_unique_picker: 
        assert l >= n 

    q = []
    while n > 0:  
        i = prg() % l
        q.append(Q[i])

        if is_unique_picker:
            Q.pop(i)
            l -= 1 
        n -= 1
    return q 


from .hop_pattern import *
from .measures import *
from copy import deepcopy


class SearchSpaceIterator:
    """
    iterator operates in a sequential/linear manner over an n-dimensional
    space with an m partition on each space. For a bounds with `n by m` specifications,
    outputs `n x m - 1` unique points in a forward (start is minumum bound point) or 
    backwards order (start is maximum bound point) 

    :param bounds: target bounds
    :type bounds: proper bounds, `m x 2`
    :param startPoint: initial point
    :type startPoint: vector, m-sized
    :param columnOrder: vector<column indices>, first is left head
    :type columnOrder: vector<int>
    :param SSIHop: hop over total distance ratio for each increment
    :type SSIHop: int
    :param cycleOn: keep going after cycle?
    :type cycleOn: bool
    :param cycleIs: cycle is maximum of bounds?
    :type cycleIs: bool
    """

    def __init__(self,bounds, startPoint, columnOrder, SSIHop = 7,cycleOn = False, cycleIs = 0):
        assert is_proper_bounds_vector(bounds), "invalid bounds"
        assert is_vector(startPoint), "invalid start point"
        assert cycleIs in [0,1], "cycle-is is wrong"

        # TODO: check column order
        self.bounds = bounds
        self.startPoint = startPoint
        self.referencePoint = np.copy(self.startPoint)
        self.cycleIs = cycleIs# 0 for left end, 1 for right end

        self.startPointContext = self.analyze_startpoint_in_bounds()
        self.columnOrder = columnOrder # alternative: could do column weights
        self.columnOrderPtr = 0  # TODO: delete, unused.
        self.hopPatterns = []

        # use the below variable if the cycle is on.
        self.cycleOn = cycleOn
        self.ssiHop = SSIHop

        self.set_hop_pattern_for_columns(self.ssiHop)

        # adjust hop pattern heads based on `cycleIs`
        self.cycleIs = cycleIs# 0 for left end, 1 for right end
        self.adjust_hop_pattern_heads()

        self.calculate_endpoint()
        self.cache = np.copy(self.startPoint)
        self.initialized = False
        return

    def de_value(self):
        return np.copy(self.referencePoint)

    def de_bounds(self):
        return np.copy(self.bounds)

    def de_start(self):
        return np.copy(self.startPoint)

    def de_end(self):
        return np.copy(self.endpoint)

    def proper_endpoint(self):
        q = []
        for (i,hp) in enumerate(self.hopPatterns):
            # calculate head
            h_ = hp.head_()
            if abs(h_ - hp.bounds[0]) < 10 ** -4:
                q.append(hp.bounds[1])
            elif abs(h_ - hp.bounds[1]) < 10 ** -4:
                q.append(hp.bounds[0])
            else:
                # make endpoint startPoint - (hopDir * 10 ** -4)
                x = 1.0 if hp.hopDirection[0] > 0 else -1.1
                q_ = self.startPoint[i] + (x * 5 * 10 ** 1)
                q.append(q_)
        return np.array(q)

    def close_cycle(self):
        """
        outputs the element after the last, and sets !terminate
        """
        pe = self.proper_endpoint()

        # turn off
        self.referencePoint = np.copy(self.endpoint)
        self.cycleOn = False

        return pe

    def adjust_hop_pattern_heads(self):
        '''
        adjusts the heads of hop pattern instances to match
        `cycleIs`
        '''

        for (i,hp) in enumerate(self.hopPatterns):
            if hp.value == self.bounds[i,0] or hp.value == self.bounds[i,1]:
                hp.head = self.bounds[i,self.cycleIs]
                hp.headIndex = self.cycleIs
            else:
                pass
        return

    def analyze_startpoint_in_bounds(self):
        """
        determines ratio of each element in startpoint with respect to
        their column size and hop partition.

        :return: vector of floats v in [0,1]
        """

        # TODO: assert start point in bounds
        v = np.asarray(self.bounds[:,1] - self.bounds[:,0], dtype=float)
        q = np.asarray(self.startPoint - self.bounds[:,0], dtype = float)
        try:
            return q / v
        except:
            return np.zeros((v.shape[0],))

    def calculate_endpoint(self):
        '''
        method called before iteration to obtain iterating endpoint
        '''

        # decrement each one
        self.endpoint = np.empty((len(self.columnOrder),))
        for (i,x) in enumerate(self.hopPatterns):
            q = deepcopy(x)
            q.rev__next__()
            self.endpoint[i] = q.rev__next__()

            # round the value to head
            q = 1

            otherV = self.bounds[i,q]
            if abs(self.endpoint[i] - otherV) < 10 ** -4.7:
                self.endpoint[i] = self.bounds[i,(q + 1) % 2]

        return np.round(self.endpoint,5)

    def set_hop_pattern_for_columns(self, dividor):
        '''
        declares n :class:`HopPattern` instances each with partition `dividor`. 

        :param dividor: partition value for each :class:`HopPattern` instance. 
        :type dividor: int
        '''
        stat = is_vector(dividor) 
        qdiv = None
        if not stat:
            assert type(dividor) in {int,np.int32,float,np.float64}
            qdiv = np.ones((self.bounds.shape[0],)) * dividor 
        else: 
            assert len(dividor) == self.bounds.shape[0]
            qdiv = dividor
        
        for (i,c) in enumerate(self.startPoint):
            diver = float(qdiv[i])
            hp = HopPattern(c, self.bounds[i,0], self.bounds[i,1], DIR = round(diver ** -1, 10))
            self.hopPatterns.append(hp)

    ###------------------------------------------------------------

    def finished(self):
        '''
        determines if iterator is finished
        
        :rtype: bool
        '''

        return not self.cycleOn and self.reached_end()

    def __next__(self):
        q = self.reached_end()
        if not self.cycleOn and q:
            #print("done with iteration")
            return np.copy(self.referencePoint)

        # check if reached end
        self.cache = np.copy(self.referencePoint)
        if q:
            self.referencePoint = np.copy(self.bounds[:,self.cycleIs])
            copi = np.copy(self.referencePoint)
        else:
            # initialize here
            copi = self.inc1()
            self.referencePoint = copi
        return np.copy(copi)

    def rev__next__(self):
        self.referencePoint = np.copy(self.rinc1())
        return np.copy(self.referencePoint)

    def rinc1(self):

        # inishiaadoe
        q = self.initiado()

        if type(q) != type(None):
            return q

        self.cache = np.empty((self.startPoint.shape[0],))

        # increment the first hop pattern
        index = len(self.columnOrder) - 1

        x = self.hopPatterns[self.columnOrder[index]].rev__next__()
        self.cache[self.columnOrder[index]] = x

        # carry-over
        index = self.carry_over(index, "finite", True)

        # for remaining indices, add value
        for i in range(index):
            self.cache[self.columnOrder[i]] = self.hopPatterns[self.columnOrder[i]].value_at()
        return np.round(self.cache, 5)

    def initiado(self):

        if not self.initialized:
            self.initialized = not self.initialized
            for i in range(len(self.hopPatterns)):
                next(self.hopPatterns[i])
            return np.copy(self.referencePoint)
        return None

    def inc1(self):
        q = self.initiado()
        if type(q) != type(None):
            return q

        # TODO: make this reference point instead
        self.cache = np.empty((self.startPoint.shape[0],))

        # increment the first hop pattern
        index = 0
        x = next(self.hopPatterns[self.columnOrder[index]])
        self.cache[self.columnOrder[index]] = x

        # carry-over
        index = self.carry_over(index, "finite")
        diff = len(self.columnOrder) - index # TODO: error here???
        for i in range(diff):
            columnOrderIndex = index + i
            columnIndex = self.columnOrder[columnOrderIndex]
            self.cache[columnIndex] = self.hopPatterns[columnIndex].value_at()
        return np.round(self.cache, 5)

    # TODO: run tests on `carryOverType`

    # TODO: rev is actually inverse
    def carry_over(self, lastIndex, carryOverType = "finite", rev = False):
        """
        method used to carry over column values to affect the next column in the
        ordering `self.columnOrder`.

        :param lastIndex: index that corresponds to column in `self.columnOrdering` where
                        the most recent delta to `self.referencePoint` occurred.
        :type lastIndex: int
        :param carryOverType: specifies if carry-over operation can cycle back to the first element in column ordering.
        :type carryOverType: str, `infinite` or `finite`.
        :param rev: carry-over in column-ordering is in decreasing order? 
        :type rev: bool
        """

        assert carryOverType in ["infinite", "finite"], "carry-over type"

        ## TODO: below two ## are for inverse mode
        if not rev:
            lf = lambda li: True if li >= len(self.columnOrder) else False
            increment = 1
        else:
            lf = lambda li: True if li < 0 else False
            increment = -1

        # check for column carry-over
        while True:
            if carryOverType == "finite" and lf(lastIndex):
                break

            modIndex = lastIndex % len(self.columnOrder)

            # check index
            if self.hopPatterns[self.columnOrder[modIndex]].did_cycle():
                modIndex2 = (modIndex + increment) % len(self.columnOrder)

                if rev:
                    x = self.hopPatterns[self.columnOrder[modIndex2]].rev__next__()
                else:
                    x = next(self.hopPatterns[self.columnOrder[modIndex2]])
                self.cache[self.columnOrder[modIndex2]] = x
            else:
                break
            lastIndex += increment

        return lastIndex

    def reached_end(self):
        '''
        determines if reference point has reached endpoint
        '''
        return equal_iterables(self.referencePoint,self.endpoint,2)

    def set_value(self,v):
        v = self.round_correct_point_in_bounds(v)
        # case: point not in bounds
        if type(v) == type(None):
            return

        self.referencePoint = v
        for (i,hp) in enumerate(self.hopPatterns):
            hp.set_value(self.referencePoint[i])

    def round_correct_point_in_bounds(self,p,roundDepth = 4):
        assert is_vector(p), "invalid vector"
        p_ = np.copy(p)
        for i in range(p_.shape[0]):
            if not (p_[i] <= self.bounds[i,1] and p_[i] >= self.bounds[i,0]):
                if abs(p_[i] - self.bounds[i,1] <= 10 ** -4):
                    p_[i] = self.bounds[i,1]
                elif abs(p_[i] - self.bounds[i,0] <= 10 ** -4):
                    p_[i] = self.bounds[i,0]
                else:
                    return None
        return p_

"""
used for improper bounds
"""
class SkewedSearchSpaceIterator(SearchSpaceIterator):
    """
    :class:`SearchSpaceIterator`-like class used to iterate over improper bounds.

    :param bounds: target bounds
    :type bounds: improper bounds, `m x 2`
    :param parentBounds: reference bounds for `self.bounds`
    :type parentBounds: proper bounds, `m x 2`
    :param startPoint: initial point
    :type startPoint: vector, m-sized
    :param columnOrder: vector<column indices>, first is left head
    :type columnOrder: vector<int>
    :param SSIHop: hop over total distance ratio for each increment
    :type SSIHop: int
    :param cycleOn: keep going after cycle?
    :type cycleOn: bool
    :param cycleIs: cycle is maximum of bounds?
    :type cycleIs: bool
    """


    def __init__(self, bounds, parentBounds,startPoint,columnOrder = None,SSIHop = 7,cycleOn = False,cycleIs = 0):
        assert not is_proper_bounds_vector(bounds), "[1] invalid bounds"
        assert is_proper_bounds_vector(parentBounds), "[2] invalid bounds"
        assert point_in_bounds(parentBounds,bounds[:,0]), "[0] invalid bound {} for {}".format(bounds[:,0],parentBounds)
        assert point_in_bounds(parentBounds,bounds[:,1]), "[1] invalid bound {} for {}".format(bounds[:,1],parentBounds)

        if type(columnOrder) == type(None):
            columnOrder = np.arange(bounds.shape[0])[::-1]

        self.referenceBounds = bounds
        self.parentBounds = parentBounds
        self.skew = self.referenceBounds[:,0] - self.parentBounds[:,0]

        # caution: no error-check
        self.sp2 = np.copy(self.referenceBounds[:,0]) if type(startPoint)\
            == type(None) else startPoint
        self.ref2 = np.copy(self.referenceBounds[:,0])

        pd = point_difference_of_improper_bounds(self.referenceBounds, self.parentBounds)
        self.iBounds = np.vstack((self.parentBounds[:,0],self.parentBounds[:,0] + pd)).T
        self.set_skew()
        sp = np.array([self.iBounds[i,self.skewEdgeIndices[i]] for i in range(self.iBounds.shape[0])])

        # declare SSI
        super().__init__(self.iBounds, np.copy(sp), columnOrder, SSIHop,cycleOn, cycleIs)

        # sort reference bounds
        self.e2 = vector_hop_in_bounds(self.endpoint,self.skew,self.parentBounds)
        return

    # TODO: test this
    @staticmethod
    def k_random_points_in_bounds(parentBounds,bounds,k):
        start = parentBounds[:,0]

        s = split_improper_bound(parentBounds,bounds,checkPoints = True)
        d2,d1 = s[1][:,1] - s[0][:,1], s[1][:,0] - s[0][:,0]
        dx = d1 + d2
        sssi = SkewedSearchSpaceIterator(bounds, parentBounds,start,None,3)

        for i in range(k):
            q = random.random() * dx
            r = start + q
            yield sssi.round_value(r)

    def de_value(self):
        return np.copy(self.referencePoint)

    def de_start(self):
        return np.copy(self.sp2)

    def de_bounds(self):
        return np.copy(self.referenceBounds)

    def de_end(self):
        return np.copy(self.e2)

    def set_skew(self):
        """
        sets skew value and its index template;
        Skew value used to calibrate value from .`self.ssi`.
        """

        skew = []
        si = []
        for i in range(self.referenceBounds.shape[0]):
            # forward traversal
            if abs(self.sp2[i] - self.referenceBounds[i,0]) <= 10 ** -5:
                skew.append(self.referenceBounds[i,0] - self.iBounds[i,0])
                si.append(0)
            # backwards traversal
            elif abs(self.sp2[i] - self.referenceBounds[i,1]) <= 10 ** -5:
                skew.append(self.referenceBounds[i,1] - self.iBounds[i,1])
                si.append(1)
            else:
                raise ValueError("cannot initialize start point at non-vertex of space")

        self.skew = np.array(skew)
        self.skewEdgeIndices = np.array(si)

    def round_value(self,v):
        return vector_hop_in_bounds(v,self.skew,self.parentBounds)

    def inverse_round_value(self,v):
        """
        inverse round value in `referenceBounds` to value in `iBounds`
        """

        return vector_hop_in_bounds(v,-self.skew,self.parentBounds)

    def __next__(self):

        if not self.cycleOn and self.reached_end():
            #print("done with iteration")
            return np.copy(self.ref2)

        q = SearchSpaceIterator.__next__(self)
        self.ref2 = self.round_value(q)
        return np.copy(self.ref2)

    def rev__next__(self):
        SearchSpaceIterator.rev__next__(self)
        self.ref2 = self.round_value(self.referencePoint)
        return np.copy(self.ref2)

    def reached_end(self):
        q = equal_iterables(self.ref2,self.e2, 4)
        return q

    def set_value(self,v):
        """
        sets value v as reference point

        :param v: vector, value in `self.referenceBounds`
        """

        assert is_vector(v), "invalid vector"
        self.ref2 = v
        q = self.inverse_round_value(v)
        SearchSpaceIterator.set_value(self,q)
        return

    def de_value(self):
        return np.copy(self.ref2)



from .search_space_iterator import *
from .relevance_functions import *
from .numerical_generator import *

class RZoom:
    '''
    RZoom is a map-like class that uses a set `rf` of relevance functions to determine if a point satisfies.
    '''

    def __init__(self, rch):
        assert type(rch) is RChainHead, "invalid RChainHead"
        self.rch = rch
        self.activationRanges = [] # each element is a bound
        self.activationRange = None
        return

    def score(self,p):
        return self.rch.apply(p)

    def output(self,p):
        s = self.score(p)
        
        # do update on activation ranges
        if s:
            if type(self.activationRange) == type(None):
                self.activationRange = np.copy(p)
            else: # assume size 1
                if len(self.activationRange.shape) != 2:
                    self.activationRange = np.vstack((self.activationRange,p)).T
                else:
                    self.activationRange[:,1] = p
            
        else:
            if type(self.activationRange) == type(None): return

            # case: 0-size bounds
            if len(self.activationRange.shape) == 1:
                self.activationRange = np.vstack((self.activationRange,\
                                    self.activationRange)).T
            self.activationRanges.append(self.activationRange)
            self.activationRange = None
        
        return

class CenterResplat:
    '''
    used for resplatting mode `png`

    :param centerBounds: proper bounds vector
    :param rch: instance<RChainHead>
    :param noiseRange: n x 2 np.ndarray
    '''

    def __init__(self,centerBounds, rch, noiseRange):
        assert is_proper_bounds_vector(centerBounds), "invalid bounds"
        self.centerBounds = centerBounds
        self.rch = rch
        self.noiseRange = noiseRange

        #: cache of relevant points passed from class<ResplattingInstructor>.
        self.relevantPointsCache = []
        #: counter for index of relevant point in `self.relevantPointsCache` to the number of times the value was selected by class<ResplattingInstructor> for re-splatting purposes.
        self.rpIndexCounter = defaultdict(int) # index

    def output(self,p):
        if self.rch.apply(p):
            self.relevantPointsCache.append(p)

    def __next__(self):
        '''
        outputs a next value based on 
        '''

        if len(self.relevantPointsCache) == 0:
            return None

        # choose random relevant point
        rpi = random.randrange(len(self.relevantPointsCache))

        self.rpIndexCounter[rpi] += 1

        v = self.relevantPointsCache[rpi]

        # add noise
        n = one_random_noise(self.centerBounds,self.noiseRange)

        # round in bounds
        return vector_hop_in_bounds(v,n,self.centerBounds)

class ResplattingInstructor:
    '''
    Class provides resplatting instructions for ResplattingSearchSpaceIterator.
    If mode is "prg", then will output random point from .CenterResplat.
    '''

    def __init__(self,rzoom,centerResplat):
        self.rzoom = rzoom
        self.rzoomBoundsCache = [] # pop(0), push(-1)
        self.centerResplat = centerResplat

    def check_args(self):
        return type(self.rzoom) == type(None) or type(self.centerResplat) == type(None)

    def output(self,p):
        assert self.check_args()
        if type(self.centerResplat) != type(None):
            self.centerResplat.output(p)
        else:
            self.rzoom.output(p)

    '''
    if mode == relevance zoom:
    '''
    def __next__(self):
        assert self.check_args()
        if type(self.centerResplat) != type(None):
            return next(self.centerResplat)
        return self.next_activation_range()

    def next_activation_range(self):
        if len(self.rzoomBoundsCache) == 0: return None
        return self.rzoomBoundsCache.pop(0)



from .ball_operator import *
from .ball_clump_data_generator import *
from .numerical_space_data_generator import *
from .message_streamer import *

# obtain the test data
ballClumpDataFilePath1 = "indep/ball_clump_data_1.txt"
ballXFilePath = "indep/ball_data.txt"


def sample_ball_x():

    q = MessageStreamer(ballXFilePath,readMode = "r")
    B = None
    c = 0
    x = []
    while q.stream():
        x.extend(q.blockData)
        c += 1

    B = Ball(x[0])
    for i in range(1,len(x)):
        B.add_element(x[i])
    return B

def pickle_obj_from_file(fp):
    fobj = open(fp,"rb")
    obj = pickle.load(fobj)
    fobj.close()
    return obj

"""
constructs a list of balls that use data generated from .BallClumpDataGenerator
"""
def sample_balls_1():

    def point_to_balls(p):
        ##print("adding point ",p)
        for (i,b) in enumerate(ballList):
            if euclidean_point_distance(b.center,p) <= radiiList[i]:
                b.add_element(p)

    def block_data_to_balls():
        for v in q.blockData:
            point_to_balls(v)

    # get the original frame of the ball clump
    fp = BallClumpDataGenerator.ballclump_frame_filepath(ballClumpDataFilePath1)
    fobj = pickle_obj_from_file(fp)
    ballList = [Ball(c) for c in fobj[1]]
    radiiList = np.round(np.array(fobj[2]),5)

    q = MessageStreamer(ballClumpDataFilePath1,readMode = "r")
    B = None
    c = 0

    while q.stream():
        block_data_to_balls()
        c += 1
    return ballList

# try splitting all the balls.



# TODO: fix import errors
from .measures import *
from .matrix_methods import *
from collections import OrderedDict

class TravelData:
    """
    Data structure that can be used to record travellistada es datas.

    :param idn: identifier for travel data
    :type idn: typically int
    :param pointData: m x 2 matrix, m is the number of samples
    :param durationData: m-iterable, timestamp data corresponding to `self.pointData`
    :param velocityData: velocity of some agent at each point in `self.pointData`
    :param hopRanges: list::((int::min, int::max), float::hop)
    :param startVelocity: initial velocity
    :type startVelocity: float
    """

    def __init__(self, idn, pointData, durationData, velocityData, hopRanges, startVelocity = None):
        self.idn = idn
        self.pointData = pointData
        self.durationData = durationData
        self.cumulativeDurationData = np.cumsum(self.durationData)
        self.velocityData = velocityData
        self.startVelocity = startVelocity
        self.accelerationData = None
        self.hopRanges = hopRanges
        self.totalDuration = np.sum(self.durationData)

    def timestamp_range(self):
        return self.cumulativeDurationData[-1] - self.cumulativeDurationData[0]

    # TODO: does not consider hopRanges, and does not
    #       use a static time unit constant.
    def velocity_to_pointwise_acceleration_data(self):
        if len(self.velocityData) == 0: raise ValueError("no velocity data for p")

        self.accelerationData = []
        prev = self.startVelocity
        for i in range(len(self.velocityData)):
            acc = self.velocityData[i] - prev
            prev = self.velocityData[i]
            self.accelerationData.append(acc)

    """
    return:
    - float
    """
    def length_of_travel(self):
        hri = 0
        travelLength = 0.0
        for (i, d) in enumerate(self.durationData):
            if not (i >= self.hopRanges[hri][0] and\
                i <= self.hopRanges[hri][1]):
                hri += 1
                ##continue
            travelLength += self.hopRanges[hri][1]
        return travelLength

    """
    return:
    - point/duration/velocity data index, previous hop increment taken
    """
    def closest_data_indices_for_wanted_duration(self, wantedDuration):

        lesser = None
        more = None
        l = len(self.cumulativeDurationData) - 1

        for i in range(l):
            if self.cumulativeDurationData[i] - wantedDuration <= 0 and\
                self.cumulativeDurationData[i + 1] - wantedDuration >= 0:
                lesser = i
                more = i + 1
                break

        if lesser == None:
            ci = l
        else:
            ci = lesser if abs(self.cumulativeDurationData[lesser] - wantedDuration)\
                < abs(self.cumulativeDurationData[more] - wantedDuration) else more

        hopIncrement = -1
        for x in self.hopRanges:
            if x[1][0] <= ci and x[1][1] >= ci:
                hopIncrement = x[0]

        assert not (hopIncrement == -1), "hop increment data for wanted duration could not be found"
        return ci, hopIncrement

    """
    """
    def velocity_to_acceleration(self, capture = False):
        a =  np.array([self.velocityData[i] - self.velocityData[i - 1]\
            for i in range(1, len(self.velocityData))])
        if capture: self.accelerationData = a
        return a

    def display(self):
        print("TRAVEL DATA")
        print("** idn: ", self.idn)
        print("** pd:\n{}\n".format(self.pointData))
        print("** dd: \n{}\n".format(self.durationData))
        print("** vd: \n{}\n".format(self.velocityData))
        print("** ad: \n{}\n".format(self.accelerationData))
        print("** hr: \n{}\n".format(self.hopRanges))

    def display_basic(self):
        print("TRAVEL DATA")
        print("** idn: ", self.idn)
        print("** first point: {}\n** last point: {}".format(self.pointData[0] if len(self.pointData) > 0 else False, self.pointData[-1] if len(self.pointData) > 0 else False))
        print("** shape of points: ", self.pointData.shape)
        print("** total duration: ", sum(self.durationData))
        print("** number of duration units: ", len(self.durationData))
        print("** number of velocity units: ", len(self.velocityData))
        print("** ending velocity: ", self.velocityData[-1] if len(self.velocityData) > 0 else False)
        print("** hr: ", self.hopRanges)

    def yield_strings_csv_format(self, stringOrList):
        assert stringOrList in [str,list], "invalid output type"
        for (p,d,v) in zip(self.pointData, self.durationData, self.velocityData):
            q = "{},{},{}".format(p,d,v) if stringOrList is str else [p,d,v]
            yield "{},{},{}".format(p,d,v)



from morebs2 import poly_interpolation
import numpy as np
import unittest

'''
python -m morebs2.tests.poly_interpolation_test  
'''
class TestLagrangePolySolverMethods(unittest.TestCase):

    def test__LagrangePolySolver_vector_form_solution(self):

        points = np.array([(1,2), (2,7), (4, -4)])
        lps = poly_interpolation.LagrangePolySolver(points)
        lps.vector_form_solution()
        xpoints = [1, 1.5, 2, 3, 3.5, 4]

        for x in xpoints:
            lb = lps.output_by_lagrange_basis(x)
            vf = lps.output_by_vector_form(x)
            self.assertTrue(abs(lb - vf) < 10 ** -5)#, "unequal values by lagrange and vector solution"
        return

    def test__LagrangePolySolver_required_velocity_for_travel_of_range_angle_effect(self):

        ## TODO:
        p3 = np.array([[0, 5],\
                    [10, 121],\
                    [16, -89],\
                    [22, 34]])

        # print lps info here
        lps = poly_interpolation.LagrangePolySolver(p3)
        rfunc = lps.continue_travelling

        # iterate through all points
        lps.form_point_sequence()
        for i in range(1,lps.data.shape[0]):
            point,prevPoint = lps.data[i], lps.data[i - 1]
            tfunc = lambda x: True if x >= point[0] else False
            v = lps.required_velocity_for_travel_of_range_angle_effect(prevPoint[0], point[0])
            dur,vel,dist,negHalt,durSat = lps.one_hop_for_travel_info(point, prevPoint, v, [], 0.0, rfunc, tfunc, "travel time")
            self.assertTrue(round(vel,3) == 0)

        return

    def test__LagrangePolySolver_required_velocity_for_travel_of_range_angle_effect_by_hop(self):

        p3 = np.array([[0, 5],\
                    [10, 121],\
                    [16, -89],\
                    [22, 34]])

        # write method to visualize poly better?
            # print lps info here
        lps = poly_interpolation.LagrangePolySolver(p3)
        rfunc = lps.continue_travelling

        v = lps.required_velocity_for_travel_of_range_angle_effect_by_hop(0, 22, 0.05)
        T,stat1,stat2 = lps.travel_from_start_for_duration_under_angle_effect(0, 100, v, 0.05)#, "travel time history")
        self.assertTrue(T.velocityData[-1] < 0.5)#, "incorrect velocity"

    def test__LagrangePolySolver_back_hop_for_velocity(self):

        p3 = np.array([[0, 5],\
                    [10, 121],\
                    [16, -89],\
                    [22, 34]])

        # write method to visualize poly better?
            # print lps info here
        lps = poly_interpolation.LagrangePolySolver(p3)
        v = lps.required_velocity_for_travel_of_range_angle_effect(21.95, 22.0)

        p1,p2 = (21.95, lps.output_by_lagrange_basis(21.95)), (22, lps.output_by_lagrange_basis(22))
        v2 =  lps.back_hop_for_velocity(p1, p2, 0.0)

        self.assertTrue(abs(v - v2) < 10 ** -5)#, "incorrect velocity for back hop"
        return

    def test__LagrangePolySolver_travel_from_start_for_duration_under_angle_effect(self):

        # case 1: non-flat polynomial, wanted duration cannot be achieved
        baselineVelocity = 25
        duration = 1
        p1 = np.array([[1,2], [2,7], [4, -4]])
        lps = poly_interpolation.LagrangePolySolver(p1)
        lps.bounds_for_x("float")

        # hop forward from .left end
        start = 1.0
        hop = 0.05

        T,stat1,stat2 = lps.travel_from_start_for_duration_under_angle_effect(start, duration, baselineVelocity, hop)#, "travel time history")
        ##T.display_basic()
        self.assertTrue(T.totalDuration <= duration)#, "[0] actual duration does not exceed wanted duration"

        # case 2: flat polynomial
        p2 = np.array([[1,2], [5,2]])
        lps = poly_interpolation.LagrangePolySolver(p2)
        lps.bounds_for_x("float")
        T,stat1,stat2 = lps.travel_from_start_for_duration_under_angle_effect(start, duration, baselineVelocity, hop)#, "travel time history")

        self.assertTrue(round(T.pointData[-1,0],5) == 5.0)#, "[1] end point {} is not 5.0".format(T.pointData[-1,0])
        self.assertTrue(np.all(np.round(T.velocityData, 3) == 25.0))#, "[1] incorrect velocity"

        # case 3: non-flat polynomial, wanted duration achieved
        p3 = np.array([[1,2], [5, 8], [13, 19], [21, 19], [34, 22]])
        lps = poly_interpolation.LagrangePolySolver(p3)
        lps.bounds_for_x("float")

        T,stat1,stat2 = lps.travel_from_start_for_duration_under_angle_effect(start, duration, baselineVelocity, hop)#, "travel time history")
        self.assertTrue(abs(T.totalDuration - 1.0) <= 10 ** -3)#, "[2] duration could not be achieved"

        # case 4: non-flat polynomial, halt restriction
        baselineVelocity = 7.0
        p4 = np.array([[1,2], [10, 9002], [20, 19002]])
        lps = poly_interpolation.LagrangePolySolver(p4)
        lps.bounds_for_x("float")

        T,stat1,stat2 = lps.travel_from_start_for_duration_under_angle_effect(start, duration, baselineVelocity, hop)#, "travel time history")
        # assert yourself


    def test__LagrangePolySolver_integral_length(self):

        # case 0: flat line for x in [2,9]
        points = np.array([[2, 6], [6,6], [9,6]])

        lps = poly_interpolation.LagrangePolySolver(points)
        lps.integral_length(2,9)
        print("integral length: ", lps.integrl)

        # polynomial cannot be straight line
        self.assertTrue(abs(lps.integrl - 7.0) < 0.25, "def. wrong integral length")

        # TODO: time-test here


    def test__LagrangePolySolver_travel_from_start_for_duration_under_angle_effect_case2(self):
        # CASE 0:
        points = np.array([[0, 5],\
                    [10, 121],\
                    [16, -89],\
                    [22, 34]])

        velocity = 10
        lps = poly_interpolation.LagrangePolySolver(points)
        T,  stat, stat2 = lps.travel_from_start_for_duration_under_angle_effect(0, 10, velocity, poly_interpolation.DEFAULT_TRAVELLING_HOP)
        self.assertTrue(T.pointData.shape[0] == 2, "case [0]: wrong number of hops")
        return

        # TODO: make assertion for below
        # CASE 1:
        points = [[200., 200.],\
            [205.51362776, 394.44128542],\
            [298.88029449, 405.49572502],\
            [337.12424895, 399.43944835],\
            [385.0893603, 261.42341642],\
            [439.36642168, 414.1732305 ]]
        points = np.array(points)
        lps = poly_interpolation.LagrangePolySolver(points)
        T,  stat, stat2 = lps.travel_from_start_for_duration_under_angle_effect(200, 15, 30, DEFAULT_TRAVELLING_HOP)#, "travel time history")
        print("T ")
        T.display_basic()
        print("\nLPS DATA")
        lps.display_basic()
        return

    def test__LagrangePolySolver_get_xrange_for_wanted_distance(self):

        # case [0]: straight line

        ##  subcase [0]
        velocity = 3
        points = np.array([[0, 4], [2,4], [7,4], [9,4]])
        lps = poly_interpolation.LagrangePolySolver(points)
        lps.bounds_for_x("float")
        e,d,stat = lps.get_xrange_for_wanted_distance(0.0, 5, integrationHop = 0.05)
        ##print('results: ', e, d, stat)
        self.assertTrue((abs(d - 5) < 0.05), "wrong value for case 0")

        ## subcase [1]: wanted distance greater than range
        e2,d2,stat2 = lps.get_xrange_for_wanted_distance(0.0, 12, integrationHop = 0.05)
        ##print('results: ', e2, d2, stat2)
        self.assertTrue(abs(d2 - 9) < 0.05, "wrong value for case 1")
        return


    def test__LagrangePolySolver_velocity_and_duration_between_points_by_angle_effect(self):

        ## CASE 0:
        p3 = np.array([[0, 5],\
                    [10, 121],\
                    [16, -89],\
                    [22, 34]])

        # write method to visualize poly better?
            # print lps info here
        lps = poly_interpolation.LagrangePolySolver(p3)

        point, prevPoint = (0.5, lps.output_by_lagrange_basis(0.5)), (0.0, lps.output_by_lagrange_basis(0.0))
        cv, travel,dur = lps.velocity_and_duration_between_points_by_angle_effect(prevPoint, point, 23, duration = 0.0, c = 0)
        self.assertTrue(not travel, "case [0]: cannot travel over range")

        # CASE 1:
        p4 = np.array([[0, 53],\
                    [16, 53]])
        velocity = 23

        lps = poly_interpolation.LagrangePolySolver(p4)
        point, prevPoint = (0.5, lps.output_by_lagrange_basis(0.5)), (0.0, lps.output_by_lagrange_basis(0.0))
        cv, travel,dur = lps.velocity_and_duration_between_points_by_angle_effect(prevPoint, point, velocity, duration = 0.0, c = 0)
        self.assertTrue(travel and abs(cv - 23.0) < 10 ** -3, "case [1]: incorrect travel results")

        # TODO: USE matplot to visualize velocity, duration variables


    def test__LagrangePolySolver__intersection_with_line(self):

        p = np.array([[120, 120], [160, 160], [250, 200]])
        lps = poly_interpolation.LagrangePolySolver(p)

        l = np.array([[130, 140], [170, 140]])
        line = poly_interpolation.Line(l)
        x,x2 = lps.intersection_with_line_v2(line)
        self.assertTrue(x2, "intersection exists b/t lps2 and l2")

    def test__LagrangePolySolver__yield_lines_clockwise_from_source_in_area(self):

        sourcePoint = np.array([50,50])
        area = np.array([[50,50], [175, 290]])

        parts = 3
        ls = poly_interpolation.LagrangePolySolver.yield_lines_clockwise_from_source_in_area(sourcePoint, area, parts)

        c = 0
        for l in ls:
            p,i = l.crosses_area_edge(area)
            self.assertTrue(i == -1, "line not in bounds")
            c += 1
        assert c == parts, "incorrect # of lines"



if __name__ == '__main__':
    unittest.main()



from setuptools import setup, find_packages

with open('README.md') as f:
    readme = f.read()

setup(
    name='morebs2',
    version='0.1.19',
    description='data structures to aid in numerical data generation and clustering',
    long_description=readme,
    author='Richard Pham',
    author_email='phamrichard45@gmail.com',
    url='https://github.com/Changissnz/morebs',
    #license=license,
    packages=find_packages(exclude=('tests','docs'))
)



from .point_sorter import *

lambda_geq = lambda x,x2: True if x >= x2 else False
lambda_leq = lambda x,x2: True if x <= x2 else False
lambda_ratio = lambda x, x2: float(x) / float(x2)
########## START: similarity measures
'''
some variations on similarity measures, functions that
measure the similarity between two vectors.
'''

def similarity_measure_cast(bounds,weights, measureType):
    assert measureType in [43,44], "invalid measure type {}".format(measureType)

    if measureType == 43:
        q = lambda v1,v2: similarity_measure__type43(v1,v2,bounds,weights)
    else:
        q = lambda v1,v2: similarity_measure__type44(v1,v2,bounds,weights)
    return q

'''
return:
- inf if `centroidValue` is `point`
  0 if inf away
'''
def similarity_measure__type1(v1,v2):
    pd = ndim_point_distance(v1,v2)
    if pd == 0: return np.inf
    pd = pd ** -1.0
    return round(pd,5)

'''
'''
def similarity_measure__type43(v1,v2, bounds, weights):
    assert len(v1) == len(v2), "unequal vector lengths"
    assert bounds.shape[0] == len(v1), "[0] invalid bounds dim"
    assert bounds.shape[1] == 2, "[1] invalid bounds dim"
    assert np.all(weights >= 0.0) and np.all(weights <= 1.0), "invalid weights"

    diff = np.abs(bounds[:,0] - bounds[:,1])

    def f_at(index):
        q = abs(v1[index] - v2[index]) / diff[index]
        return q * weights[index]

    k = len(v1)
    s = 0.0
    for i in range(k):
        s += f_at(i)
    return round(s / k, 5) if k != 0 else 0.0

def similarity_measure__type44(v1,v2,bounds,weights):
    assert len(v1) == len(v2), "unequal vector lengths"
    assert bounds.shape[0] == len(v1), "[0] invalid bounds dim"
    assert bounds.shape[1] == 2, "[1] invalid bounds dim"
    assert weights.shape[0] == len(v1), "[0] invalid weights dim"
    assert weights.shape[1] == 2, "[1] invalid weights dim"

    #
    def vector_sign(v):
        if np.all(v >= 0.0):
            return 1
        elif np.all(v <= 1.0):
            return -1
        else:
            return 0

    def half_of_value(v, b, defaultHalf = None):
        assert len(b) == 2, "invalid range"
        assert defaultHalf in [-1,1,None], "default half is incorrect"
        m = (b[1] - b[0]) / 2.0
        if v > m: return 1
        elif v < m: return -1
        else: return defaultHalf

    vs1 = vector_sign(weights[:,0])
    vs2 = vector_sign(weights[:,1])

    # special assertions for weight signs
    assert vs1 != 0 and vs2 != 0, "invalid column signs for weights"
    assert vs1 != vs2, "must have opposing signs for weights"

    def f_score(v_, index):
        assert v_ in [1,2], "invalid v"
        q = v1 if v_ == 1 else v2
        hf = half_of_value(q[index], bounds[index],None)

        if hf == 1:
            return 2.0 * abs(bounds[index,hf] - q[index]) / abs(bounds[index,0] - bounds[index,1])
        elif hf == -1:
            return 2.0 * abs(bounds[index,hf + 1] - q[index]) / abs(bounds[index,0] - bounds[index,1])
        else:
            return 0

    def f_at(index):
        # get the two f-scores
        fs1 = f_score(1,index)
        fs2 = f_score(2,index)

        # multiply by weights
        fs1 = fs1 * weights[index,0]
        fs2 = fs2 * weights[index,1]

        return abs(fs1 - fs2)

    k = len(v1)
    s = 0.0
    for i in range(k):
        s += f_at(i)

    return round(s / (k * 2.0),5) if k != 0 else 0.0


'''
description:
- similarity measure considers radius info of centroid that v1 belongs to.
  Points `v2` that are farther away from the centroid C(v1) will have higher score.
'''
def similarity_measure__type45(v1,v2, meanRadius,minRadius,maxRadius):

    # perform calculation involving radiusInfo
    # get ratio of mean point to min and max point

    # ratio > 1 => mean radius is closest to max radius\
    if maxRadius - minRadius != 0:
        ratio = meanRadius / float (maxRadius - minRadius)

    pd = similarity_measure__type1(v1,v2)
    return round(pd * ratio,5)

############################ TODO: future
'''
description:
- memory-aware of last significant point score

centroidValue := vector, must be equal in length to `point`
point := vector, point to measure the weight of
lastSignificantPointScore := float
moreSignificant := 1 to use (>=), 0 to use (<=)

return:
- float, `lastSignificantPointScore` if score S of `point` passes `pwf` else S
'''
def point_weight_function_3(centroidValue, point, lastSignificantPointScore, pwf, scoreCompFunc):

    s = pwf(centroidValue,point)
    if scoreCompFunc(lastSignificantPointScore,s):
        # replacement value is lastSignificantPointScore
        s = lastSignificantPointScore
    return s



from .ball_clump_data_generator import *

def sample_args__ball_clump_frame_generator1():

    bounds = np.array([[-100,100.0],\
                    [-100,100.0],\
                    [-100,100.0],\
                    [-100,100.0],\
                    [-100,100.0],\
                    [-100,100.0]])

    # case 1
    ballRadii = [10.0,40.0,50.0,100.0,10.0,30.0,15.0]

    c1  = [[0.25,0.4,0.55],\
            [0.1],\
            [0.2,0.2,0.2,0.2],\
            [0.5, 0.5],
            [0.3,0.3,0.2],\
            [0.1,0.25,0.3],\
            [0.05,0.4]]

    c2 = [[1000,100,500],\
            [300],\
            [250,450,180,220],\
            [2000,4000],\
            [2000,2000,300],\
            [500,400,1800],\
            [100,200]]

    cs = (c1,c2)
    filePath = "indep/ball_clump_data_1.txt"
    return bounds,ballRadii,cs,filePath

def test__BallClumpDataGenerator__sample1():
    bounds,ballRadii,cs,filePath = sample_args__ball_clump_frame_generator1()

    bcdg = BallClumpDataGenerator(bounds,ballRadii,cs,filePath)
    bcdg.set_frame()

    for b in bcdg.balls:
        print(b)
        print()

    ssz = BallClumpDataGenerator.ball_subset_size_in_bounds(bcdg.rbounds,bcdg.balls)
    ##print("SSZ ",ssz)
    assert BallClumpDataGenerator.ball_set_in_bounds(bcdg.rbounds,bcdg.balls), "invalid balls generated"
    print("* after setting frame")
    bcdg.make_data()
    bcdg.save_frame()


def test__BallClumpDataGenerator__ballclump_frame_filepath():
    fp = "indep/ball_clump_data_1.txt"
    fp2 = BallClumpDataGenerator.ballclump_frame_filepath(fp)
    assert fp2 == "indep/ball_clump_data_1__frame.txt"

    fp = "ball_clump_data_1.txt"
    fp2 = BallClumpDataGenerator.ballclump_frame_filepath(fp)
    assert fp2 == "ball_clump_data_1__frame.txt"
    return



from morebs2 import deline
import numpy as np

import unittest

'''
python -m morebs2.tests.deline_test  
'''
class DelineClass(unittest.TestCase):

    def test__DLineate22__collect_break_points__case_1(self):
        data = deline.test_dataset__Dlineate22_1()
        dl = deline.DLineate22(data)
        dl.preprocess()
        dl.collect_break_points()

        l = np.array([[ 5.,  0.],\
        [5.,6.],\
        [ 5.,  9.],\
        [ 5., 12.],\
        [ 5., 15.]])
        
        r = np.array([[20., 15.],\
        [15., 12.],\
        [15.,  9.],\
        [25, 7.5],\
        [20.,  0.]])

        t = np.array([[ 5. , 15.],\
        [15. , 12.],\
        [20. , 15.],\
        [25. ,  7.5]])

        b = np.array([[25. ,  7.5],\
        [20. ,  0.],\
        [15. ,  9.],\
        [ 5. ,  0.]])

        assert np.all(dl.d.d['l'] == l),"got {}".format(dl.d.d['l'])
        assert np.all(dl.d.d['r'] == r),"got {}".format(dl.d.d['r'])
        assert np.all(dl.d.d['t'] == t),"got {}".format(dl.d.d['t'])
        assert np.all(dl.d.d['b'] == b),"got {}".format(dl.d.d['b'])

    def test__DLineate22__collect_break_points__case_2(self):
        # clockwise 
        data = deline.test_dataset__Dlineate22_2()
        dl = deline.DLineate22(data)
        dl.preprocess()
        dl.collect_break_points()

        l = np.array([[ 5.,  0.],\
            [ 5., 15.]])

        r = np.array([[20., 15.],\
            [20.,  0.]])

        t = np.array([[ 5., 15.],\
            [20., 15.]])

        b = np.array([[20.,  0.],\
            [ 5.,  0.]])

        assert np.all(dl.d.d['l'] == l),"got {}".format(dl.d.d['l'])
        assert np.all(dl.d.d['r'] == r),"got {}".format(dl.d.d['r'])
        assert np.all(dl.d.d['t'] == t),"got {}".format(dl.d.d['t'])
        assert np.all(dl.d.d['b'] == b),"got {}".format(dl.d.d['b'])

        # counter-clockwise 
        data = deline.test_dataset__Dlineate22_2()
        dl = deline.DLineate22(data,False)
        dl.preprocess()
        dl.collect_break_points()

        l = np.array([[ 5., 15.],\
            [ 5.,  0.]])

        r = np.array([[20.,  0.],\
            [20., 15.]])

        t = np.array([[20., 15.],\
            [ 5., 15.]])

        b = np.array([[ 5.,  0.],\
            [20.,  0.]])

        assert np.all(dl.d.d['l'] == l),"got {}".format(dl.d.d['l'])
        assert np.all(dl.d.d['r'] == r),"got {}".format(dl.d.d['r'])
        assert np.all(dl.d.d['t'] == t),"got {}".format(dl.d.d['t'])
        assert np.all(dl.d.d['b'] == b),"got {}".format(dl.d.d['b'])

    def test__DLineate22__classify_points__case_3(self):
        td3 = deline.test_dataset__Dlineate22_3()
        p = [12.5,7.5]

        dl = deline.DLineate22(np.copy(td3),dmethod="nocross")
        dl.preprocess()
        dl.collect_break_points()

        c = dl.d.classify_point(p)    
        assert c == 0.0, "dmethod=nocross incorrectly classifies"

        dl = deline.DLineate22(np.copy(td3),dmethod="nodup")
        dl.preprocess()
        dl.collect_break_points()

        c = dl.d.classify_point(p)    
        assert c == 0.0, "dmethod=nodup incorrectly classifies"

        dl = deline.DLineate22(np.copy(td3),dmethod="nojag")
        dl.preprocess()
        dl.collect_break_points()

        c = dl.d.classify_point(p)    
        assert c == 0.0, "dmethod=nojag incorrectly classifies"

    def test__DLineate22__classify_points__case_4(self):
        td4 = deline.test_dataset__Dlineate22_4()
        dl = deline.DLineate22(np.copy(td4),dmethod="nodup")
        dl.preprocess()
        dl.collect_break_points()

        p = [12.5,7.5]
        c = dl.d.classify_point(p)
        assert c == 0, "misclassification for case 4, dmethod=nodup"

        dl = deline.DLineate22(np.copy(td4),dmethod="nocross")
        dl.preprocess()
        dl.collect_break_points()

        p = [12.5,7.5]
        c = dl.d.classify_point(p)
        assert c == 0, "misclassification for case 4, dmethod=nocross"

        dl = deline.DLineate22(np.copy(td4),dmethod="nojag")
        dl.preprocess()
        dl.collect_break_points()

        p = [12.5,7.5]
        c = dl.d.classify_point(p)
        assert c == 0, "misclassification for case 4, dmethod=nojag"

        p = [25,7.5]
        c = dl.d.classify_point(p)
        assert c == -1, "misclassification for case 4, dmethod=nojag"

    def test__DLineate22__classify_points__case_4(self):
        data = deline.test_dataset__DLineateMC_1()

        dl = deline.DLineate22(data,dmethod="nojag")
        dl.preprocess()
        dl.collect_break_points()

        c = dl.d.classify_point([65,20.])
        assert c == 1., "incorrect classification for point 1, got {} want {}".format(c,1)

        c = dl.d.classify_point([80,20.])
        assert c == 1., "incorrect classification for point 2, got {} want {}".format(c,1)
        return
    
    def test__DLineate22__collect_break_points__AND__classify_point__case_1(self):
        data = deline.test_dataset__Dlineate22_1_v2()
        dl = deline.DLineate22(data)
        dl.preprocess()
        dl.collect_break_points()

        for q in dl.lpoints:
            c = dl.d.classify_point(q)
            assert c == 0
     
    def test__DLineate22__optimize_delineation__case_1(self):
        data = deline.test_dataset__Dlineate22_1_v3(500,[0.000001,1.])
        dl = deline.DLineate22(data)
        dl.preprocess()
        dl.collect_break_points()
        s = dl.optimize_delineation()
        assert s > 0, "optimization should reduce classification error!"
    

if __name__ == '__main__':
    unittest.main()